/* ***********************************************************************/
/*                                                                       */
/*                      WiNRADiO G31DDC EXCALIBUR                        */
/*                      G31DDC device driver (USB)                       */
/*                     (C) Copyright 2015 RADIXON.                       */
/*                                                                       */
/* ***********************************************************************/

#include "g31ddc.h"

static void g31_io_buffer_unmap(struct g31_io_buffer *io_buffer)
{
 int i;

#ifdef __G31_USB__
    if(io_buffer->mem!=NULL)
    {
        vunmap(io_buffer->mem);
        io_buffer->mem=NULL;
    }
#endif //__G31_USB__

    if(io_buffer->pages)
    {
        if(io_buffer->changed)
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                SetPageDirty(io_buffer->pages[i]);
                put_page(io_buffer->pages[i]);
            }
        }
        else
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                put_page(io_buffer->pages[i]);
            }
        }

        io_buffer->changed=0;

        kfree(io_buffer->pages);
        io_buffer->pages=NULL;
        io_buffer->page_count=0;
    }
}


static int g31_io_buffer_prepare(uintptr_t user_buffer,uint32_t size,int write,struct g31_io_buffer **io_buffer_r)
{
 struct g31_io_buffer *io_buffer;
 int page_count,locked_page_count,i;

    if(unlikely(user_buffer==0 || (user_buffer & (~PAGE_MASK))!=0 || size==0))
    {
        return -EINVAL;
    }

    page_count=(int)(((user_buffer+size-1)&PAGE_MASK)>>PAGE_SHIFT)-(int)(user_buffer>>PAGE_SHIFT)+1;

    io_buffer=kmalloc(sizeof(struct g31_io_buffer),GFP_KERNEL);

    if(unlikely(io_buffer==NULL))
    {
        return -ENOMEM;
    }

    io_buffer->pages=kmalloc(page_count*sizeof(struct page*),GFP_KERNEL);

    if(unlikely(io_buffer->pages==NULL))
    {
        kfree(io_buffer);
        return -ENOMEM;
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
    locked_page_count=get_user_pages_fast((uintptr_t)user_buffer,page_count,write,io_buffer->pages);
#else
    down_read(&current->mm->mmap_sem);
    locked_page_count=get_user_pages(current,current->mm,(uintptr_t)user_buffer,page_count,write,1,io_buffer->pages,NULL);
    up_read(&current->mm->mmap_sem);
#endif

    if(unlikely(locked_page_count<0))
    {
        kfree(io_buffer->pages);
        kfree(io_buffer);
        return locked_page_count;
    }

    if(unlikely(locked_page_count<page_count))
    {
        for(i=0;i<locked_page_count;i++)
        {
            put_page(io_buffer->pages[i]);
        }

        kfree(io_buffer->pages);
        kfree(io_buffer);
        return -ENOMEM;
    }

    io_buffer->user_buffer=user_buffer;
    io_buffer->size=size;
    io_buffer->page_count=locked_page_count;
    io_buffer->changed=0;

#ifdef __G31_USB__
    io_buffer->mem=NULL;
#endif //__G31_USB__

#ifdef __G31_PCI__
    io_buffer->sg_list=NULL;
    io_buffer->sg_segment_count=0;
    io_buffer->sg_dev=NULL;
#endif //__G31_PCI__

    *io_buffer_r=io_buffer;

    return 0;
}

#ifdef __G31_USB__

static int g31_io_buffer_map(struct g31_io_buffer *io_buffer)
{
    if(unlikely(io_buffer->mem!=NULL))
    {
        return 0;
    }

    io_buffer->mem=vmap(io_buffer->pages,io_buffer->page_count,0,PAGE_KERNEL);

    if(unlikely(io_buffer->mem==NULL))
    {
        return -ENOMEM;
    }

    return 0;
}

#endif //__G31_USB__

#ifdef __G31_PCI__

inline static void g31_io_buffer_sg_sync_for_cpu(struct g31_io_buffer *io_buffer)
{
    dma_sync_sg_for_cpu(io_buffer->sg_dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
}

inline static void g31_io_buffer_sg_sync_for_device(struct g31_io_buffer *io_buffer)
{
    dma_sync_sg_for_device(io_buffer->sg_dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
}

static int g31_io_buffer_sg_map(struct g31_io_buffer *io_buffer,struct device *dev,enum dma_data_direction direction)
{
 struct scatterlist *sg_list;
 int i;
 int count;
 uint32_t size;

    sg_list=kcalloc(io_buffer->page_count,sizeof(*sg_list),GFP_KERNEL);

    if(unlikely(sg_list==NULL))
    {
        return -ENOMEM;
    }

    size=io_buffer->size;

    for(i=0;i<io_buffer->page_count;i++)
    {
        if(likely(size>=PAGE_SIZE))
        {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
            sg_set_page(&sg_list[i],io_buffer->pages[i],PAGE_SIZE,0);
#else
            sg_list[i].page=io_buffer->pages[i];
            sg_list[i].offset=0;
            sg_list[i].length=PAGE_SIZE;
#endif

            size-=PAGE_SIZE;
        }
        else
        {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
            sg_set_page(&sg_list[i],io_buffer->pages[i],size,0);
#else
            sg_list[i].page=io_buffer->pages[i];
            sg_list[i].offset=0;
            sg_list[i].length=size;
#endif

            size=0;
        }
    }

    count=dma_map_sg(dev,sg_list,io_buffer->page_count,direction);

    if(unlikely(count==0))
    {
        kfree(sg_list);
        return -ENOMEM;
    }

    io_buffer->sg_list=sg_list;
    io_buffer->sg_segment_count=count;
    io_buffer->sg_dev=dev;
    io_buffer->sg_direction=direction;

    return 0;
}

static void g31_io_buffer_sg_unmap(struct g31_io_buffer *io_buffer)
{
    if(io_buffer->sg_list)
    {
        g31_io_buffer_sg_sync_for_cpu(io_buffer);

        dma_unmap_sg(io_buffer->sg_dev,io_buffer->sg_list,io_buffer->page_count,io_buffer->sg_direction);
        kfree(io_buffer->sg_list);

        io_buffer->sg_list=NULL;
    }
}

#endif //__G31_PCI__

static void g31_io_buffer_free(struct g31_io_buffer *io_buffer)
{
 int i;

#ifdef __G31_USB__
    if(io_buffer->mem!=NULL)
    {
        vunmap(io_buffer->mem);
    }
#endif //__G31_USB__

#ifdef __G31_PCI__
    g31_io_buffer_sg_unmap(io_buffer);
#endif //__G31_PCI__

    if(io_buffer->pages!=NULL)
    {
        if(io_buffer->changed)
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                if(!PageReserved(io_buffer->pages[i]))
                {
                    SetPageDirty(io_buffer->pages[i]);
                }

                put_page(io_buffer->pages[i]);
            }
        }
        else
        {
            for(i=0;i<io_buffer->page_count;i++)
            {
                put_page(io_buffer->pages[i]);
            }
        }

        kfree(io_buffer->pages);
    }

    kfree(io_buffer);
}

static int g31_io_stream_get_complete_buffer(struct g31ddc *dev,struct g31_io_stream *io_stream,uintptr_t *user_buffer)
{
 struct g31_io_buffer *io_buffer;
 unsigned long flags;
 struct g31_complete_io_buffers *buffers;

    buffers=&io_stream->complete_buffers;

    for(;;)
    {
        down_read(&io_stream->rw_semaphore);

        if(unlikely(!io_stream->active))
        {
            up_read(&io_stream->rw_semaphore);
            return -EPERM;
        }

        spin_lock_irqsave(&buffers->spinlock,flags);

        if(unlikely(list_empty(&buffers->list)))
        {
            io_buffer=NULL;
        }
        else
        {
            io_buffer=list_entry(buffers->list.next,struct g31_io_buffer,list_entry);

            list_del(&io_buffer->list_entry);
        }

        spin_unlock_irqrestore(&buffers->spinlock,flags);

        up_read(&io_stream->rw_semaphore);

        if(likely(io_buffer!=NULL))
        {
            break;
        }

        if(unlikely(down_interruptible(&buffers->semaphore)<0))
        {
            return -EINTR;
        }
    }

    *user_buffer=io_buffer->user_buffer;

    g31_io_buffer_free(io_buffer);

    return 0;
}

#ifdef __G31_USB__
static struct g31_io_buffer* g31_io_stream_get_pending_buffer(struct g31_io_stream *io_stream)
{
 unsigned long flags;
 struct g31_pending_io_buffers *buffers;
 struct g31_io_buffer *io_buffer;

    buffers=&io_stream->pending_buffers;

    spin_lock_irqsave(&buffers->spinlock,flags);

    if(unlikely(list_empty(&buffers->list)))
    {
        io_buffer=NULL;
    }
    else
    {
        io_buffer=list_entry(buffers->list.next,struct g31_io_buffer,list_entry);

        list_del(buffers->list.next);
    }

    spin_unlock_irqrestore(&buffers->spinlock,flags);

    return io_buffer;
}
#endif
#ifdef __G31_PCI__

static struct g31_io_buffer* g31_io_stream_get_pending_buffer_nolock(struct g31_io_stream *io_stream)
{
 struct g31_pending_io_buffers *buffers;
 struct g31_io_buffer *io_buffer;

    buffers=&io_stream->pending_buffers;

    if(unlikely(list_empty(&buffers->list)))
    {
        io_buffer=NULL;
    }
    else
    {
        io_buffer=list_entry(buffers->list.next,struct g31_io_buffer,list_entry);

        list_del(buffers->list.next);
    }

    return io_buffer;
}

#endif //__G31_PCI__

static void g31_io_stream_put_complete_buffer(struct g31_io_stream *io_stream,struct g31_io_buffer *io_buffer)
{
 unsigned long flags;
 struct g31_complete_io_buffers *buffers;

    buffers=&io_stream->complete_buffers;

    g31_io_buffer_unmap(io_buffer);

    spin_lock_irqsave(&buffers->spinlock,flags);

    list_add_tail(&io_buffer->list_entry,&buffers->list);

    up(&buffers->semaphore);

    spin_unlock_irqrestore(&buffers->spinlock,flags);
}


#ifdef __G31_USB__

static void g31_io_stream_put_pending_buffer(struct g31_io_stream *io_stream,struct g31_io_buffer *io_buffer)
{
 unsigned long flags;
 struct g31_pending_io_buffers *buffers;

    buffers=&io_stream->pending_buffers;

    spin_lock_irqsave(&buffers->spinlock,flags);

    list_add_tail(&io_buffer->list_entry,&buffers->list);

    spin_unlock_irqrestore(&buffers->spinlock,flags);
}

static int g31_io_stream_put_pending_buffer2(struct g31_io_stream *io_stream,struct g31_io_buffer *io_buffer)
{
 unsigned long flags;
 struct g31_pending_io_buffers *buffers;
 int first;

    buffers=&io_stream->pending_buffers;

    spin_lock_irqsave(&buffers->spinlock,flags);

    first=list_empty(&buffers->list);

    list_add_tail(&io_buffer->list_entry,&buffers->list);

    spin_unlock_irqrestore(&buffers->spinlock,flags);

    return first;
}

#endif //__G31_USB__ 
inline static void g31_io_stream_put_pending_buffer_nolock(struct g31_io_stream *io_stream,struct g31_io_buffer *io_buffer)
{
    list_add_tail(&io_buffer->list_entry,&io_stream->pending_buffers.list);
}

static void g31_io_stream_free_buffers(struct g31_io_stream *io_stream)
{
 struct g31_io_buffer *io_buffer;

    while(!list_empty(&io_stream->pending_buffers.list))
    {
        io_buffer=list_entry(io_stream->pending_buffers.list.next,struct g31_io_buffer,list_entry);

        list_del(&io_buffer->list_entry);

        g31_io_buffer_free(io_buffer);
    }

    INIT_LIST_HEAD(&io_stream->pending_buffers.list);

    while(!list_empty(&io_stream->complete_buffers.list))
    {
        io_buffer=list_entry(io_stream->complete_buffers.list.next,struct g31_io_buffer,list_entry);

        list_del(&io_buffer->list_entry);

        g31_io_buffer_free(io_buffer);
    }

    INIT_LIST_HEAD(&io_stream->complete_buffers.list);
}

static dev_t g31_dev0;

static struct class g31_class=
{
    .name=G31_DEVICE_NAME
};

static struct g31ddc *g31_registered_devices[G31_MAX_DEVICE_COUNT]={0};

static struct g31_mutex g31_reg_mutex;

static int g31_set_power(struct g31ddc *dev,uint8_t power,uint8_t lock);

static int g31_fo_open(struct inode *node, struct file *file);
static int g31_fo_release(struct inode *node, struct file *file);
static int g31_fo_mmap(struct file *file,struct vm_area_struct *vma);

#ifdef HAVE_UNLOCKED_IOCTL
static long g31_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
#else
static int g31_fo_ioctl(struct inode *node,struct file *file, unsigned int cmd, unsigned long arg);
#endif

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
static long g31_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg);
#endif

static struct file_operations g31_file_operations=
{
    .open=g31_fo_open,
    .release=g31_fo_release,
#ifdef HAVE_UNLOCKED_IOCTL
    .unlocked_ioctl=g31_fo_ioctl,
#else
    .ioctl=g31_fo_ioctl,
#endif //HAVE_UNLOCKED_IOCTL

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
    .compat_ioctl=g31_fo_ioctl_compat,
#endif //defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
    .mmap=g31_fo_mmap,
};

static void g31_wd_timer(unsigned long arg)
{
 struct g31ddc *dev;

    dev=(struct g31ddc*)arg;

    queue_work(dev->workqueue,&dev->wd_work);
}

static void g31_wd_work_handler(struct work_struct *work)
{
 struct g31ddc *dev;
 int status;

    dev=container_of(work,struct g31ddc,wd_work);

    G31_DEV_LOCK;

    if(likely(dev->state.open))
    {
        if(dev->state.power)
        {
            status=dev->func_wd(dev);

            if(likely(status==0))
            {
                dev->wd_counter=0;
            }
            else if(status>0)
            {
                dev->wd_counter++;

                if(unlikely(dev->wd_counter>=24))
                {
                    dev->state.temperature=1;
                    g31_set_power(dev,0,0);
                    dev->wd_counter=0;
                }
            }
        }

        dev->wd_timer.expires=jiffies+2*HZ;
        add_timer(&dev->wd_timer);
    }

    G31_DEV_UNLOCK;
}

static void g31_initialize(struct g31ddc *dev)
{
 int i;

    dev->state.connected=1;
    g31_mutex_init(&dev->mutex);
    init_waitqueue_head(&dev->power_wait_queue);

    dev->last_power_off_time=current_kernel_time();

    for(i=0;i<G31_IO_STREAM_COUNT;i++)
    {
        init_rwsem(&dev->io_stream[i].rw_semaphore);

        INIT_LIST_HEAD(&dev->io_stream[i].complete_buffers.list);
        sema_init(&dev->io_stream[i].complete_buffers.semaphore,0);
        spin_lock_init(&dev->io_stream[i].complete_buffers.spinlock);

        INIT_LIST_HEAD(&dev->io_stream[i].pending_buffers.list);
        spin_lock_init(&dev->io_stream[i].pending_buffers.spinlock);
    }
    
    INIT_WORK(&dev->wd_work,g31_wd_work_handler);

    init_timer(&dev->wd_timer);

    dev->wd_timer.function=g31_wd_timer;
    dev->wd_timer.data=(unsigned long)dev;
}

static void g31_uninitialize(struct g31ddc *dev)
{
    g31_mutex_destroy(&dev->mutex);
}

static void g31_delete(struct g31ddc *dev)
{
    g31_set_power(dev,0,1);

    dev->func_delete(dev);

    g31_uninitialize(dev);

    kfree(dev);
}

static void g31_release_callback(struct device *dev)
{
 struct g31ddc *g31ddc;

    g31ddc=container_of(dev,struct g31ddc,dev);

    g31_delete(g31ddc);
}

static int g31_register(struct g31ddc *dev)
{
int status;
int index;
dev_t dev_no;

    G31_REG_LOCK;

    for(index=0;index<G31_MAX_DEVICE_COUNT;index++)
    {
        if(g31_registered_devices[index]==NULL)
        {
            break;
        }
    }

    if(index>=G31_MAX_DEVICE_COUNT)
    {
        G31_REG_UNLOCK;

        G31_ERROR("no free device slot available!\n");
        return -ENFILE;
    }

    dev->index=index;

    dev_no=MKDEV(MAJOR(g31_dev0),index);

    cdev_init(&dev->cdev,&g31_file_operations);

    dev->cdev.owner=THIS_MODULE;

    status=cdev_add(&dev->cdev,dev_no,1);

    if(unlikely(status<0))
    {
        G31_REG_UNLOCK;
        G31_ERROR("char device registration failed!\n");
        return status;
    }

    memset(&dev->dev,0,sizeof(dev->dev));

    dev->dev.devt=dev_no;
    dev->dev.class=&g31_class;
    dev->dev.release=g31_release_callback;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)
    dev_set_name(&dev->dev,G31_DEVICE_NAME"%d",index);
#else
    snprintf(dev->dev.bus_id,sizeof(dev->dev.bus_id),G31_DEVICE_NAME"%d",index);
#endif

    status=device_register(&dev->dev);

    if(unlikely(status<0))
    {
        cdev_del(&dev->cdev);
        G31_REG_UNLOCK;
        G31_ERROR("failed to register device!\n");
        return status;
    }

    g31_registered_devices[index]=dev;

    G31_REG_UNLOCK;

    return 0;
}

static void g31_unregister(struct g31ddc *dev)
{
    G31_REG_LOCK;

    if(likely(g31_registered_devices[dev->index]==dev))
    {
        g31_registered_devices[dev->index]=NULL;

        G31_REG_UNLOCK;

        cdev_del(&dev->cdev);
        device_unregister(&dev->dev);
    }
    else
    {
        G31_REG_UNLOCK;
    }
}

static int g31_set_power(struct g31ddc *dev,uint8_t power,uint8_t lock)
{
 int status;
 struct timespec time;
 int i;

    power=power?1:0;

    if(lock)
    {
        G31_DEV_LOCK;
    }

    if(dev->state.power==power)
    {
        status=0;
    }
    else
    {
        if(power)
        {
            dev->state.suspended=0;
            dev->state.temperature=0;

            time=current_kernel_time();
            time.tv_nsec-=500000000;

            if(time.tv_nsec<0)
            {
                time.tv_sec--;
                time.tv_nsec+=1000000000;
            }

            if((dev->last_power_off_time.tv_sec==0 && dev->last_power_off_time.tv_nsec==0) ||
               timespec_compare(&time,&dev->last_power_off_time)<0)
            {
                msleep(500);
            }
        }
        else
        {
            dev->func_if_stop(dev);

            for(i=0;i<G31_DDC1_COUNT;i++)
            {
                dev->func_ddc1_stop(dev,i);
            }
        }

        status=dev->func_set_power(dev,power);

        if(likely(status>=0))
        {
            if(power)
            {
                dev->state.power=1;
                msleep(100);
            }
            else
            {
                dev->state.power=0;
                dev->state.fpga=0;
                dev->last_power_off_time=current_kernel_time();

                wake_up_all(&dev->power_wait_queue);
            }
        }
    }

    if(lock)
    {
        G31_DEV_UNLOCK;
    }

return status;
}

static int g31_configure_fpga(struct g31ddc *dev,void __user *user_buffer,uint32_t size)
{
 int status;

    if(unlikely(size==0 || user_buffer==NULL))
    {
        return -EINVAL;
    }

    if(unlikely(!access_ok(VERIFY_READ,user_buffer,size)))
    {
        return -EFAULT;
    }

    G31_DEV_LOCK;

    if(unlikely(!dev->state.power))
    {
        G31_DEV_UNLOCK;
        return -EPERM;
    }

    status=dev->func_configure_fpga(dev,user_buffer,size);

    if(likely(status>=0))
    {
        dev->state.fpga=1;

        msleep(50);
    }
    else
    {
        G31_ERROR("failed to initialize FPGA!\n");
    }

    G31_DEV_UNLOCK;

    return status;
}
/*
static int g31_set_spi_regs(struct g31ddc *dev,struct g31_spi_register __user *user_buffer,uint16_t count,uint16_t id)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_set_spi_regs(dev,user_buffer,count,id);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}

static int g31_set_spi2_regs(struct g31ddc *dev,struct g31_spi2_register __user *user_buffer,uint16_t count,uint16_t id)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_set_spi2_regs(dev,user_buffer,count,id);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}
*/

static int g31_rf_control1(struct g31ddc *dev,struct g31_rf_control *rf)
{
int status;

G31_DEV_LOCK;

if(likely(dev->state.fpga)) status=dev->func_rf_control1(dev,rf);
else status=-EPERM;

G31_DEV_UNLOCK;

return status;
}
//*
static int g31_rf_control2(struct g31ddc *dev,struct g31_rf_control2 *rf2)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_rf_control2(dev,rf2);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}
//*/
static int g31_set_noise_blanker(struct g31ddc *dev,uint16_t value)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_set_noise_blanker(dev,value);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}

static int g31_if_start(struct g31ddc *dev,uint16_t period)
{
 int status;

    if(unlikely(period==0))
    {
        return -EINVAL;
    }

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        if(unlikely(dev->io_stream[G31_IO_STREAM_IN_IF].active ))
        {
            status=-EBUSY;
        }
        else
        {
            status=dev->func_if_start(dev,period);
        }
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}

static void g31_if_stop(struct g31ddc *dev)
{
    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        dev->func_if_stop(dev);
    }

    G31_DEV_UNLOCK;
}
/*
static int g31_get_temprature(struct g31ddc *dev,uint32_t *temp)
{
 int status;
    if(unlikely(!access_ok(VERIFY_WRITE,(void __user*)temp,sizeof(uint32_t))))
    {
        return -EFAULT;
    }
status=dev->func_get_temp(dev,temp);
return status;
}
//*/
inline static int g31_get_complete_buffer(struct g31ddc *dev,uint32_t stream_index,unsigned long arg)
{
 uintptr_t user_buffer;
 int status;

    if(unlikely(stream_index>=G31_IO_STREAM_COUNT))
    {
        return -EINVAL;
    }

    if(unlikely(!access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(void*))))
    {
        return -EFAULT;
    }

    status=g31_io_stream_get_complete_buffer(dev,&dev->io_stream[stream_index],&user_buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return __put_user(user_buffer,(uintptr_t __user*)arg);
}

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)

inline static int g31_get_complete_buffer_compat(struct g31ddc *dev,uint32_t stream_index,unsigned long arg)
{
 uintptr_t user_buffer;
 int status;

    if(unlikely(!access_ok(VERIFY_WRITE,(void __user*)compat_ptr(arg),sizeof(compat_uptr_t))))
    {
        return -EFAULT;
    }

    status=g31_io_stream_get_complete_buffer(dev,&dev->io_stream[stream_index],&user_buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return __put_user(ptr_to_compat((void __user*)user_buffer),(compat_uptr_t __user*)compat_ptr(arg));
}

#endif //defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)


static int g31_put_buffer(struct g31ddc *dev,uint32_t stream_index,void __user *user_buffer,int write)
{
 struct g31_io_buffer *io_buffer;
 struct g31_io_stream *io_stream;
 int status;

    io_stream=dev->io_stream+stream_index;

    down_read(&io_stream->rw_semaphore);

    if(unlikely(!io_stream->active))
    {
        up_read(&io_stream->rw_semaphore);
        return -EPERM;
    }

    status=g31_io_buffer_prepare((uintptr_t)user_buffer,io_stream->buffer_size,write,&io_buffer);

    if(unlikely(status))
    {
        up_read(&io_stream->rw_semaphore);
        return status;
    }

    status=dev->func_put_buffer(dev,stream_index,io_buffer);

    up_read(&io_stream->rw_semaphore);

    if(unlikely(status<0))
    {
        g31_io_buffer_free(io_buffer);
    }

    return status;
}

static int g31_get_adc_level(struct g31ddc *dev,uint16_t *level)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_get_adc_level(dev,level);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}

static int g31_get_adc_sample_rate(struct g31ddc *dev,uint32_t *sr)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        status=dev->func_get_adc_sample_rate(dev,sr);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}

static int g31_ddc1_start(struct g31ddc *dev,uint32_t channel,struct g31_ddc1_params *params,uint32_t __user *user_coeffs,uint32_t buffer_size,uint32_t bytes_per_sec)
{
 int status;
 uint32_t *coeffs;
 struct g31_io_stream *io_stream;

    if(unlikely(buffer_size>G31_MAX_BUFFER_SIZE))
    {
        return -E2BIG;
    }

    coeffs=kmalloc(params->fir_len*sizeof(uint32_t)/2,GFP_KERNEL);

    if(unlikely(coeffs==NULL))
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(coeffs,user_coeffs,params->fir_len*sizeof(uint32_t)/2)!=0))
    {
        kfree(coeffs);
        return -EFAULT;
    }

    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        io_stream=dev->io_stream+G31_IO_STREAM_IN_DDC+channel;

        if(unlikely(io_stream->active))
        {
            status=-EBUSY;
        }
        else
        {
            io_stream->buffer_size=buffer_size;
            io_stream->bytes_per_sec=bytes_per_sec;

            dev->ddc1_params[channel]=*params;

            status=dev->func_ddc1_start(dev,channel,coeffs);
        }
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    kfree(coeffs);

    return status;
}

static void g31_ddc1_stop(struct g31ddc *dev,uint32_t channel)
{
    G31_DEV_LOCK;

    if(likely(dev->state.fpga))
    {
        dev->func_ddc1_stop(dev,channel);
    }

    G31_DEV_UNLOCK;
}


static int g31_ddc1_set_frequency(struct g31ddc *dev,uint32_t channel,uint32_t tuning_word)
{
 int status;

    G31_DEV_LOCK;

    if(likely(dev->state.fpga && dev->io_stream[G31_IO_STREAM_IN_DDC+channel].active))
    {
        dev->ddc1_params[channel].tuning_word=tuning_word;

        status=dev->func_ddc1_update_params(dev,channel);
    }
    else
    {
        status=-EPERM;
    }

    G31_DEV_UNLOCK;

    return status;
}


static void g31_cancel_buffers(struct g31ddc *dev,int stream_index)
{
 struct g31_io_stream *io_stream;

    io_stream=&dev->io_stream[stream_index];

    down_read(&io_stream->rw_semaphore);

    if(io_stream->active)
    {
        dev->func_cancel_buffers(dev,stream_index);
    }

    up_read(&io_stream->rw_semaphore);
}

#ifdef __G31_USB__

static int g31_usb_initialize(struct g31ddc_usb *dev);

uint16_t DeviceUSB_GetNextIFCommandId(struct g31ddc_usb *dev)
{
if(dev->IFCommandId>=0x7FFF) return (dev->IFCommandId=0x0000);
else return ++dev->IFCommandId;
}

static int g31_usb_fpga_cmd(struct g31ddc_usb *dev,uint16_t target,uint16_t id,uint16_t *data,uint32_t size)
{
uint32_t i;
uint16_t c;
int sent;
uint16_t packet[64];

if(unlikely(size>(64-9)*sizeof(uint16_t))) return -E2BIG;

memset(packet,0,sizeof(packet));

packet[0]=cpu_to_le16(0x4041);
packet[1]=cpu_to_le16(0x4243);
packet[2]=cpu_to_le16(0x4445);
packet[3]=cpu_to_le16(0x4647);
packet[4]=cpu_to_le16(target);
packet[5]=cpu_to_le16(id);
//packet[6]=0;
packet[7]=cpu_to_le16(10000);

memcpy(packet+9,data,size);

c=0;
for(i=0;i<64;i++) c+=le16_to_cpu(packet[i]);
c=(~c)+1;
packet[8]=cpu_to_le16(c);

return usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),packet,sizeof(packet),&sent,G31_USB_TIMEOUT);
}

static int g31_usb_rf_control1(struct g31ddc_usb *dev,struct g31_rf_control *rf)
{
uint16_t Packet[4];

uint16_t Att=(rf->Atten/3)&7;

Packet[0]=0x01;
Packet[1]=Att>>2;

if(rf->WettingCurrent) Packet[1]|=0x02;
//if(!rf->Preamp) Packet[1]|=0x04;

Packet[2]=(rf->PreselIndexHigh>>1)|(Att<<14);
Packet[3]=rf->PreselIndexLow|(rf->PreselIndexHigh<<15);

//DEBUG(DbgPrint(DRIVER_BASE_NAME":DeviceUSB_SendSPIPacket Packet: %04X %04X %04X",Packet[0],Packet[1],Packet[2]));

return g31_usb_fpga_cmd(dev,3,DeviceUSB_GetNextIFCommandId(dev),Packet,sizeof(Packet));
}
//*
static int g31_usb_rf_control2(struct g31ddc_usb *dev,struct g31_rf_control2 *rf2)
{
uint16_t Packet=0;
if (rf2->Spreading) Packet=0x01;
if (rf2->Dither) Packet|=0x02;
//if(rf2->ExternalReference) Packet|=0x04;

return g31_usb_fpga_cmd(dev,2,DeviceUSB_GetNextIFCommandId(dev),&Packet,sizeof(Packet));
}
//*/
static int g31_usb_setup(struct g31ddc_usb *dev)
{
 int i,num_endpoints,cnt;
 int status;
 struct usb_endpoint_descriptor *ep_desc;


    status=usb_set_interface(dev->usb_dev,0,0);

    if(unlikely(status<0))
    {
        G31_ERROR("failed to set interface and alternate number!\n");
        return status;
    }

    num_endpoints=dev->usb_interface->cur_altsetting->desc.bNumEndpoints;

    cnt=0;

    for(i=0;i<num_endpoints;i++)
    {
        ep_desc=&dev->usb_interface->cur_altsetting->endpoint[i].desc;

//		G31_ERROR("EP #%02x: %d bytes\n",(int)ep_desc->bEndpointAddress,(int)ep_desc->wMaxPacketSize);

        switch(ep_desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK)
        {
            case G31_USB_EP_OUT_CTRL:
/*
                if(le16_to_cpu(ep_desc->wMaxPacketSize)==512)
                {
                    ep_desc->wMaxPacketSize=cpu_to_le16(128);
                }
                else if(le16_to_cpu(ep_desc->wMaxPacketSize)==255)
                {
                    ep_desc->wMaxPacketSize=cpu_to_le16(128);
                }
//*/
            case G31_USB_EP_IN_FPGA:
            case G31_USB_EP_IN_CTRL:
            case G31_USB_EP_OUT_FPGA:
                cnt++;
                break;
        }

        status=usb_control_msg(dev->usb_dev, usb_sndctrlpipe(dev->usb_dev, 0),
                              USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                              USB_ENDPOINT_HALT,ep_desc->bEndpointAddress, NULL, 0,
                              USB_CTRL_SET_TIMEOUT);
        if (status<0) {
			G31_ERROR("EP %d HALT error %d",(int)ep_desc->bEndpointAddress,status);
			}
    }
/*
	status=usb_control_msg(dev->usb_dev, usb_sndctrlpipe(dev->usb_dev, 0),USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,USB_ENDPOINT_HALT,0,NULL,0,USB_CTRL_SET_TIMEOUT);
	if (status<0) {
		G31_ERROR("EP 0snd HALT error %d",status);
		}
		
	status=usb_control_msg(dev->usb_dev, usb_rcvctrlpipe(dev->usb_dev, 0),USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,USB_ENDPOINT_HALT,0,NULL,0,USB_CTRL_SET_TIMEOUT);
	if (status<0) {
		G31_ERROR("EP 0rcv HALT error %d",status);
		}
//*/
    if(unlikely(cnt!=4))
    {
       G31_ERROR("required enpoints not found!\n");
       return -ENODEV;
    }
/*		// toto ma len G39DDC
    status=usb_control_msg(dev->usb_dev, usb_sndctrlpipe(dev->usb_dev,0),0x0F,USB_TYPE_VENDOR|USB_DIR_OUT,0,0, NULL,0,G31_USB_TIMEOUT);
    if (status<0) {
			G31_ERROR("CP test error %d",status);
			}
//*/
    return 0;
}


static void g31_usb_delete(struct g31ddc_usb *dev)
{
    usb_put_intf(dev->usb_interface);
    usb_put_dev(dev->usb_dev);
}

static int g31_usb_open(struct g31ddc_usb *dev)
{
    return 0;
}



static void g31_usb_close(struct g31ddc_usb *dev)
{
 int len;
 uint8_t packet[12]={0x04,0x00,0x00,0x00,0x0D,0x00,0x00,0x00,0xA8,0x00,0x00,0x00};

    usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),packet,sizeof(packet),&len,G31_USB_TIMEOUT);
}


static int g31_usb_control1_read(struct g31ddc_usb *dev,const uint16_t *params,void __user *user_buffer,uint16_t buffer_size,void __user *transferred)
{
 int status;
 void *buffer;

    if(unlikely(transferred!=NULL && !access_ok(transferred,VERIFY_WRITE,sizeof(uint16_t))))
    {
        return -EFAULT;
    }

    if(buffer_size!=0)
    {
        if(!access_ok(user_buffer,VERIFY_WRITE,buffer_size))
        {
            return -EFAULT;
        }

        buffer=kmalloc(buffer_size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }
    }
    else
    {
        buffer=NULL;
    }

    status=usb_control_msg(dev->usb_dev,usb_rcvctrlpipe(dev->usb_dev,0),
                params[0],USB_TYPE_VENDOR|USB_DIR_IN,
                params[1],params[2],buffer,buffer_size,G31_USB_TIMEOUT);

    if(likely(status>=0))
    {
        if(transferred!=NULL)
        {
            if(unlikely(__put_user(status,(uint16_t __user*)transferred)<0))
            {
                if(buffer!=NULL)
                {
                    kfree(buffer);
                }
                return -EFAULT;
            }
        }

        if(buffer!=NULL)
        {
            if(status>0)
            {
                if(unlikely(__copy_to_user(user_buffer,buffer,status)!=0))
                {
                    kfree(buffer);
                    return -EFAULT;
                }
            }

            kfree(buffer);
        }

        return 0;
    }
    else
    {
        return status;
    }
}

static int g31_usb_control1_write(struct g31ddc_usb *dev,const uint16_t *params,void __user *user_buffer,uint16_t buffer_size,void __user *transferred)
{
 int status;
 void *buffer;

    if(unlikely(transferred!=NULL && !access_ok(transferred,VERIFY_WRITE,sizeof(uint16_t))))
    {
        return -EFAULT;
    }

    if(buffer_size!=0)
    {
        buffer=kmalloc(buffer_size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }

        if(unlikely(copy_from_user(buffer,user_buffer,buffer_size)!=0))
        {
            kfree(buffer);
            return -EFAULT;
        }
    }
    else
    {
        buffer=NULL;
    }

    status=usb_control_msg(dev->usb_dev,usb_sndctrlpipe(dev->usb_dev,0),
                params[0],USB_TYPE_VENDOR|USB_DIR_OUT,
                params[1],params[2],buffer,buffer_size,G31_USB_TIMEOUT);

    if(buffer!=NULL)
    {
        kfree(buffer);
    }

    if(likely(status>=0))
    {
        if(transferred!=NULL)
        {
            if(unlikely(__put_user(status,(uint16_t __user*)transferred)<0))
            {
                return -EFAULT;
            }
        }

        return 0;
    }
    else
    {
        return status;
    }
}

static int g31_usb_control2_read(struct g31ddc_usb *dev,void __user *user_buffer,uint16_t buffer_size,void __user *transferred)
{
 int status;
 int len;
 void *buffer;

    if(unlikely(transferred!=NULL && !access_ok(transferred,VERIFY_WRITE,sizeof(uint16_t))))
    {
        return -EFAULT;
    }

    if(buffer_size!=0)
    {
        if(!access_ok(user_buffer,VERIFY_WRITE,buffer_size))
        {
            return -EFAULT;
        }

        buffer=kmalloc(buffer_size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }
    }
    else
    {
        buffer=NULL;
    }

    status=usb_bulk_msg(dev->usb_dev,usb_rcvbulkpipe(dev->usb_dev,G31_USB_EP_IN_CTRL),buffer,buffer_size,&len,G31_USB_TIMEOUT);

    if(likely(status>=0))
    {
        if(transferred!=NULL)
        {
            if(unlikely(__put_user(len,(uint16_t __user*)transferred)<0))
            {
                if(buffer!=NULL)
                {
                    kfree(buffer);
                }
                return -EFAULT;
            }
        }

        if(buffer!=NULL)
        {
            if(len>0)
            {
                if(unlikely(__copy_to_user(user_buffer,buffer,len)!=0))
                {
                    kfree(buffer);
                    return -EFAULT;
                }
            }

            kfree(buffer);
        }

        return 0;
    }
    else
    {
        if(buffer!=NULL)
        {
            kfree(buffer);
        }

        return status;
    }
}


static int g31_usb_control2_write(struct g31ddc_usb *dev,void __user *user_buffer,uint16_t buffer_size,void __user *transferred)
{
 int status;
 int len;
 void *buffer;

    if(unlikely(transferred!=NULL && !access_ok(transferred,VERIFY_WRITE,sizeof(uint16_t))))
    {
        return -EFAULT;
    }

    if(buffer_size!=0)
    {
        buffer=kmalloc(buffer_size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }

        if(unlikely(copy_from_user(buffer,user_buffer,buffer_size)!=0))
        {
            kfree(buffer);
            return -EFAULT;
        }
    }
    else
    {
        buffer=NULL;
    }

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),buffer,buffer_size,&len,G31_USB_TIMEOUT);

    if(buffer!=NULL)
    {
        kfree(buffer);
    }

    if(unlikely(status<0))
    {
        return status;
    }

    if(transferred!=NULL)
    {
        if(unlikely(__put_user(len,(uint16_t __user *)transferred)<0))
        {
            return -EFAULT;
        }
    }

    return 0;
}

static int g31_usb_set_power(struct g31ddc_usb *dev,uint8_t power)
{
int len;
uint8_t packet[12]={0x04,0x00,0x00,0x00,0x11,0x00,0x00,0x00,power?0x03:0x01,0x00,0x00,0x00};

return usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),packet,sizeof(packet),&len,G31_USB_TIMEOUT);
}

#define FPGA_CHUNK 256

static int g31_usb_configure_fpga(struct g31ddc_usb *dev,void __user *user_buffer,uint32_t size)
{
 void *buffer;
 int status;
 int bytes_to_send;
 int len;
 uint32_t packet[3];

// void __user *usrbuff=user_buffer;
// int cnt=0;
//	G31_ERROR("FPGA boot %p [%u]\n",user_buffer,size);

    buffer=kmalloc(FPGA_CHUNK,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
		G31_ERROR("Failed to allocate memory for FPGA boot\n");
        return -ENOMEM;
    }

    packet[0]=cpu_to_le32(size);
    packet[1]=cpu_to_le32(0x10);

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),packet,8,&len,G31_USB_TIMEOUT);

    if(unlikely(status<0))
    {
		G31_ERROR("Failed to init FPGA boot, status=%d\n",status);
        kfree(buffer);
        return status;
    }

    while(size!=0)
    {
        if(size>=FPGA_CHUNK) bytes_to_send=FPGA_CHUNK;
        else bytes_to_send=size;

        if(unlikely(copy_from_user(buffer,user_buffer,bytes_to_send)!=0))
        {
			G31_ERROR("Failed to copy the FPGA boot\n");
            kfree(buffer);
            return -EFAULT;
        }

        status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),buffer,bytes_to_send,&len,G31_USB_TIMEOUT);

        if(unlikely(status<0))
        {
			G31_ERROR("FPGA config BULK msg failed, %d -> %d, len=%d, status=%d\n",bytes_to_send,size,len,status);
            kfree(buffer);
            return status;
        }
//        else G31_ERROR("FPGA config BULK send OK, %d -> %d, len=%d, status=%d\n",bytes_to_send,size,len,status);

        size-=len;
        user_buffer=(void __user*)((uintptr_t)user_buffer+len);
    }

    kfree(buffer);

    msleep(50);

	len=0;
    status=usb_bulk_msg(dev->usb_dev,usb_rcvbulkpipe(dev->usb_dev,G31_USB_EP_IN_CTRL),packet,12,&len,G31_USB_TIMEOUT);

    if(unlikely(status<0))
    {
		G31_ERROR("Failed to finish FPGA boot, len=%d, status=%d\n",len,status);
        return -EIO;
    }

    if(unlikely(len!=12 || le32_to_cpu(packet[0])!=4 || le32_to_cpu(packet[1])!=0x80000010 || le32_to_cpu(packet[2])!=0))
    {
		G31_ERROR("FPGA boot failed\n");
        return -EIO;
    }

    return 0;
}

//*/
/*
static int g31_usb_set_spi_regs(struct g31ddc_usb *dev,struct g31_spi2_register __user *user_buffer,uint16_t count,uint16_t id)
{
 uint32_t packet_size,i;
 uint32_t buffer_size;
 int status;
 uint16_t *packet;
 struct g31_spi_register *buffer,*data;

    buffer_size=count*sizeof(struct g31_spi_register);

    buffer=kmalloc(buffer_size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(buffer,user_buffer,buffer_size)!=0))
    {
        kfree(buffer);
        return -EFAULT;
    }

    packet_size=buffer_size+sizeof(uint16_t);

    packet=kmalloc(packet_size,GFP_KERNEL);

    if(unlikely(packet==NULL))
    {
        kfree(buffer);
        return -ENOMEM;
    }

    packet[0]=cpu_to_le32(count);

    data=(struct g31_spi_register*)(packet+1);

    for(i=0;i<count;i++)
    {
        data[i].address=cpu_to_le16(buffer[i].address);
        data[i].value=cpu_to_le32(buffer[i].value);
    }

    status=g31_usb_fpga_cmd(dev,3,id,packet,packet_size);

    kfree(packet);
    kfree(buffer);

    return status;
}
//*/
/*
static int g31_usb_set_spi2_regs(struct g31ddc_usb *dev,struct g31_spi2_register __user *user_buffer,uint16_t count,uint16_t id)
{
 uint32_t size;
 uint32_t i;
 uint16_t *packet,*data;
 int status;
 int sent;
 struct g31_spi2_register *buffer;

    size=count*sizeof(struct g31_spi2_register);

    buffer=kmalloc(size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }


    if(unlikely(copy_from_user(buffer,user_buffer,size)!=0))
    {
        kfree(buffer);
        return -EFAULT;
    }

    size=(count*6+3)*sizeof(uint16_t);

    packet=kmalloc(size,GFP_KERNEL);

    if(unlikely(packet==NULL))
    {
        kfree(buffer);
        return -ENOMEM;
    }

    packet[0]=cpu_to_le16(0x4041);
    packet[1]=cpu_to_le16(3+count*6);
    packet[2]=cpu_to_le16(1);

    data=packet+3;

    for(i=0;i<count;i++)
    {
        data[0]=cpu_to_le16(4);
        *(uint32_t*)(data+1)=cpu_to_le32(buffer[i].u32[0]);
        data[3]=cpu_to_le16(5);
        *(uint32_t*)(data+4)=cpu_to_le32(buffer[i].u32[1]);
        data+=6;
    }

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),packet,size,&sent,G31_USB_TIMEOUT);

    kfree(packet);
    kfree(buffer);

    return status;
}
//*/
/*
static int g31_usb_set_reg(struct g31ddc_usb *dev,uint16_t address,uint32_t value)
{
 int sent;
 int status;
 uint16_t *packet;

    packet=kmalloc(6*sizeof(uint16_t),GFP_KERNEL);

    if(unlikely(packet==NULL))
    {
        return -ENOMEM;
    }

    packet[0]=cpu_to_le16(0x4041);
    packet[1]=cpu_to_le16(6);
    packet[2]=cpu_to_le16(1);
    packet[3]=cpu_to_le16(address);
    *(uint32_t*)(packet+4)=cpu_to_le32(value);

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),packet,6*sizeof(uint16_t),&sent,G31_USB_TIMEOUT);

    kfree(packet);

    return status;
}

static int g31_usb_set_reg_rep(struct g31ddc_usb *dev,uint16_t address,uint32_t *values,uint32_t count)
{
 int sent;
 uint32_t i,size;
 uint32_t *data;
 uint16_t *packet;
 int status;

    size=count*sizeof(uint32_t)+4*sizeof(uint16_t);

    packet=kmalloc(size,GFP_KERNEL);

    if(unlikely(packet==NULL))
    {
        return -ENOMEM;
    }

    packet[0]=cpu_to_le16(0x4041);
    packet[1]=cpu_to_le16(2*count+4);
    packet[2]=cpu_to_le16(2);
    packet[3]=cpu_to_le16(address);

    data=(uint32_t*)(packet+4);

    for(i=0;i<count;i++)
    {
        data[i]=cpu_to_le32(values[i]);
    }

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),packet,size,&sent,G31_USB_TIMEOUT);

    kfree(packet);

    return status;
}

static int g31_usb_set_regs(struct g31ddc_usb *dev,uint16_t *address,uint32_t *value,uint32_t count)
{
 int sent;
 int status;
 uint32_t size,i;
 uint16_t *data,*packet;

    size=3*sizeof(uint16_t)*(count+1);

    packet=kmalloc(size,GFP_KERNEL);

    if(unlikely(packet==NULL))
    {
        return -ENOMEM;
    }

    packet[0]=cpu_to_le16(0x4041);
    packet[1]=cpu_to_le16(3+3*count);
    packet[2]=cpu_to_le16(1);

    data=packet+3;

    for(i=0;i<count;i++,data+=3)
    {
        data[0]=cpu_to_le16(address[i]);
        data[1]=cpu_to_le16((uint16_t)((value[i])&0xFFFF));
        data[2]=cpu_to_le16((uint16_t)((value[i])>>16));
    }

    status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),packet,size,&sent,G31_USB_TIMEOUT);

    kfree(packet);

    return status;
}
//*/
static int g31_usb_set_noise_blanker(struct g31ddc_usb *dev,uint16_t value)
{
return g31_usb_fpga_cmd(dev,8,DeviceUSB_GetNextIFCommandId(dev),&value,sizeof(value));
}

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
static void g31_usb_in_urb_complete_callback(struct urb *urb)
#else
static void g31_usb_in_urb_complete_callback(struct urb *urb,struct pt_regs *regs)
#endif
{
 struct g31ddc_usb *dev;
 struct g31_usb_buffer *buffer;
 unsigned long flags;

    buffer=urb->context;
    dev=buffer->dev;

    spin_lock_irqsave(&dev->input.complete_usb_buffers.spinlock,flags);

    buffer->pending=0;

    list_add_tail(&buffer->list_entry,&dev->input.complete_usb_buffers.list);

    spin_unlock_irqrestore(&dev->input.complete_usb_buffers.spinlock,flags);

    wake_up_process(dev->input.thread);
}


static void g31_usb_update_input_buffer_size(struct g31ddc_usb *dev)
{
 uint32_t bytes_per_sec;
 uint32_t buffer_size;
 int i;

    bytes_per_sec=0;

    for(i=G31_IO_STREAM_IN_FIRST;i<=G31_IO_STREAM_IN_LAST;i++)
    {
        if(dev->g31_dev.io_stream[i].active)
        {
            bytes_per_sec+=dev->g31_dev.io_stream[i].bytes_per_sec;
        }
    }

    bytes_per_sec+=bytes_per_sec*6/G31_USB_INPUT_PACKET_PAYLOAD_SIZE;

    buffer_size=(bytes_per_sec*10/1000+2047)&(~2047);

    if(buffer_size<G31_USB_INPUT_BUFFER_MIN_SIZE)
    {
        buffer_size=G31_USB_INPUT_BUFFER_MIN_SIZE;
    }
    else if(buffer_size>G31_USB_INPUT_BUFFER_MAX_SIZE)
    {
        buffer_size=G31_USB_INPUT_BUFFER_MAX_SIZE;
    }

    dev->input.buffer_size=buffer_size;
}

static int g31_usb_input_stream_thread(void *param)
{
 struct g31ddc_usb *dev=param;
 int i;
 int status;
 int loop;
 uint32_t rest_buffer_size;
 struct g31_usb_buffer *usb_buffer;
 struct g31_usb_buffers *usb_buffers;
 struct g31_usb_input_packet *packet;
 uint32_t stream_index;
 uint32_t size;
 unsigned long flags;
 struct g31_io_stream *io_stream;
 struct g31_io_buffer **io_buffers;
 uint32_t remain;
 uint8_t *packet_data;
 uint32_t transferred[G31_IO_STREAM_COUNT]={0,0};
 uint8_t rest_buffer[sizeof(struct g31_usb_input_packet)];

    rest_buffer_size=0;
    usb_buffers=&dev->input.complete_usb_buffers;
    loop=1;

    for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
    {
        usb_fill_bulk_urb(dev->input.buffer[i].urb,dev->usb_dev,
                          usb_rcvbulkpipe(dev->usb_dev,G31_USB_EP_IN_FPGA),
                          dev->input.buffer[i].mem+sizeof(struct g31_usb_input_packet),
                          dev->input.buffer_size,g31_usb_in_urb_complete_callback,
                          dev->input.buffer+i);

        dev->input.buffer[i].pending=1;

        status=usb_submit_urb(dev->input.buffer[i].urb,GFP_KERNEL);

        if(unlikely(status<0))
        {
            dev->input.buffer[i].pending=0;
            break;
        }
    }

    if(likely(status>=0))
    {
        io_buffers=dev->io_buffer;

        while(loop && !kthread_should_stop())
        {
            spin_lock_irqsave(&usb_buffers->spinlock,flags);

            if(unlikely(list_empty(&usb_buffers->list)))
            {
                usb_buffer=NULL;
            }
            else
            {
                usb_buffer=list_entry(usb_buffers->list.next,struct g31_usb_buffer,list_entry);
                list_del(&usb_buffer->list_entry);
            }

            spin_unlock_irqrestore(&usb_buffers->spinlock,flags);

            if(likely(usb_buffer!=NULL))
            {
                if(likely(usb_buffer->urb->status==0 && usb_buffer->urb->actual_length>0))
                {
                    size=usb_buffer->urb->actual_length;

                    if(rest_buffer_size!=0)
                    {
                        packet=(struct g31_usb_input_packet*)(usb_buffer->mem+sizeof(struct g31_usb_input_packet)-rest_buffer_size);
                        memcpy(packet,rest_buffer,rest_buffer_size);
                        size+=rest_buffer_size;
                    }
                    else
                    {
                        packet=(struct g31_usb_input_packet*)(usb_buffer->mem+sizeof(struct g31_usb_input_packet));
                    }

                    while(loop && size>=sizeof(struct g31_usb_input_packet))
                    {
                        if(unlikely(packet->sync[0]!=0x41 || packet->sync[1]!=0x40))
                        {
							int skipcnt=size;
                            do
                            {
                                size--;
                                packet=(struct g31_usb_input_packet*)(((uintptr_t)packet)+1);
                            }
                            while(size>=2 && (packet->sync[0]!=0x41 || packet->sync[1]!=0x40));
                            skipcnt-=size;
                            G31_ERROR("Skipped %d bytes in the stream\n",skipcnt);

                            if(size<sizeof(struct g31_usb_input_packet))
                            {
                                break;
                            }
                        }
						
                        switch(packet->source)
                        {
/*
                            case 1: //DDC1[0]
                            case 2: //DDC1[1]
                            case 3: //DDC2[0]
//*/
                            case 6: //DDC2[1]
                                stream_index=G31_IO_STREAM_IN_DDC;
                                io_stream=&dev->g31_dev.io_stream[stream_index];

                                down_read(&io_stream->rw_semaphore);

                                if (io_stream->active)// && dev->IFCommandId==le16_to_cpu(packet->id))
                                {
                                    i=G31_USB_INPUT_PACKET_PAYLOAD_SIZE;
                                    packet_data=packet->data;

                                    while(i)
                                    {
                                        if(unlikely(io_buffers[stream_index]==NULL))
                                        {
                                            io_buffers[stream_index]=g31_io_stream_get_pending_buffer(io_stream);

                                            if(unlikely(io_buffers[stream_index]==NULL))
                                            {
                                                break;
                                            }

                                            if(unlikely(g31_io_buffer_map(io_buffers[stream_index])<0))
                                            {
                                                g31_io_buffer_free(io_buffers[stream_index]);
                                                io_buffers[stream_index]=NULL;
                                                loop=0;
                                                break;
                                            }

                                            io_buffers[stream_index]->changed=1;
                                            transferred[stream_index]=0;
                                        }

                                        remain=io_stream->buffer_size-transferred[stream_index];

                                        if(i>=remain)
                                        {
//											G31_ERROR("Complete buffer\n");
                                            memcpy(io_buffers[stream_index]->mem+transferred[stream_index],packet_data,remain);

                                            g31_io_stream_put_complete_buffer(io_stream,io_buffers[stream_index]);
                                            io_buffers[stream_index]=NULL;

                                            i-=remain;
                                            packet_data+=remain;
                                        }
                                        else
                                        {
//											G31_ERROR("Uncomplete buffer %d+%d/%d\n",transferred[stream_index],i,io_stream->buffer_size);
                                            memcpy(io_buffers[stream_index]->mem+transferred[stream_index],packet_data,i);
                                            transferred[stream_index]+=i;
                                            break;
                                        }
                                    }
                                }

                                up_read(&io_stream->rw_semaphore);
                                break;

                            case 4: //IF
                                io_stream=&dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF];

                                down_read(&io_stream->rw_semaphore);

                                if(io_stream->active /*&& dev->IFCommandId==le16_to_cpu(packet->id)*/)
                                {
                                    if(packet->flags & 0x80)
                                    {
                                        if(likely(io_buffers[G31_IO_STREAM_IN_IF]==NULL))
                                        {
                                            io_buffers[G31_IO_STREAM_IN_IF]=g31_io_stream_get_pending_buffer(io_stream);

                                            if(unlikely(io_buffers[G31_IO_STREAM_IN_IF]==NULL))
                                            {
                                                up_read(&io_stream->rw_semaphore);
                                                break;
                                            }

                                            if(unlikely(g31_io_buffer_map(io_buffers[G31_IO_STREAM_IN_IF])<0))
                                            {
                                                g31_io_buffer_free(io_buffers[G31_IO_STREAM_IN_IF]);
                                                io_buffers[G31_IO_STREAM_IN_IF]=NULL;

                                                up_read(&io_stream->rw_semaphore);
                                                loop=0;
                                                break;
                                            }

                                            io_buffers[G31_IO_STREAM_IN_IF]->changed=1;
                                            transferred[G31_IO_STREAM_IN_IF]=0;
                                        }

                                        memcpy(io_buffers[G31_IO_STREAM_IN_IF]->mem,packet->data+6,G31_USB_INPUT_PACKET_PAYLOAD_SIZE-6);
                                        transferred[G31_IO_STREAM_IN_IF]=G31_USB_INPUT_PACKET_PAYLOAD_SIZE-6;

                                        dev->ADC_Level=packet->data[0]+(((uint16_t)packet->data[1])<<8);
                                        dev->TCXO_Freq=((uint32_t)packet->data[2]<<16)+((uint32_t)packet->data[3]<<24)+((uint32_t)packet->data[4]<<8)+((uint32_t)packet->data[5]);
                                        if (!dev->TCXO_Freq) dev->TCXO_Freq=10000000;
                                    }
                                    else
                                    {
                                        if(likely(dev->io_buffer[G31_IO_STREAM_IN_IF]!=NULL))
                                        {
                                            remain=io_stream->buffer_size-transferred[G31_IO_STREAM_IN_IF];

                                            if(remain<=G31_USB_INPUT_PACKET_PAYLOAD_SIZE)
                                            {
                                                memcpy(io_buffers[G31_IO_STREAM_IN_IF]->mem+
                                                       transferred[G31_IO_STREAM_IN_IF],
                                                       packet->data,remain);
                                                transferred[G31_IO_STREAM_IN_IF]+=remain;

                                                g31_io_stream_put_complete_buffer(io_stream,io_buffers[G31_IO_STREAM_IN_IF]);

                                                io_buffers[G31_IO_STREAM_IN_IF]=NULL;
                                            }
                                            else
                                            {
                                                memcpy(io_buffers[G31_IO_STREAM_IN_IF]->mem+
                                                       transferred[G31_IO_STREAM_IN_IF],
                                                       packet->data,G31_USB_INPUT_PACKET_PAYLOAD_SIZE);
                                                transferred[G31_IO_STREAM_IN_IF]+=G31_USB_INPUT_PACKET_PAYLOAD_SIZE;
                                            }
                                        }
                                    }
                                }

                                up_read(&io_stream->rw_semaphore);
                                break;
                        }

                        packet++;
                        size-=sizeof(struct g31_usb_input_packet);
                    }

                    rest_buffer_size=size;

                    if(size!=0)
                    {
                        memcpy(rest_buffer,packet,size);
                    }
                }

                if(likely(loop))
                {
                    usb_fill_bulk_urb(usb_buffer->urb,dev->usb_dev,
                                      usb_rcvbulkpipe(dev->usb_dev,G31_USB_EP_IN_FPGA),
                                      usb_buffer->mem+sizeof(struct g31_usb_input_packet),
                                      dev->input.buffer_size,g31_usb_in_urb_complete_callback,
                                      usb_buffer);

                    usb_buffer->pending=1;

                    status=usb_submit_urb(usb_buffer->urb,GFP_KERNEL);

                    if(unlikely(status<0))
                    {
                        usb_buffer->pending=0;
                        break;
                    }
                }
            }
            else
            {
                set_current_state(TASK_INTERRUPTIBLE);

                schedule();

                set_current_state(TASK_RUNNING);
            }
        }
    }

    for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
    {
        if(dev->input.buffer[i].pending)
        {
            usb_kill_urb(dev->input.buffer[i].urb);
        }
    }


    while(!kthread_should_stop())
    {
        set_current_state(TASK_INTERRUPTIBLE);

        schedule();

        set_current_state(TASK_RUNNING);
    }

    return 0;
}


/*
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
static void g31_usb_out_urb_complete_callback(struct urb *urb)
#else
static void g31_usb_out_urb_complete_callback(struct urb *urb,struct pt_regs *regs)
#endif
{
 struct g31ddc_usb *dev;
 struct g31_usb_buffer *buffer;
 unsigned long flags;

    buffer=urb->context;
    dev=buffer->dev;

    spin_lock_irqsave(&dev->output.avail_usb_buffers.spinlock,flags);

    buffer->pending=0;

    list_add_tail(&buffer->list_entry,&dev->output.avail_usb_buffers.list);

    spin_unlock_irqrestore(&dev->output.avail_usb_buffers.spinlock,flags);

    wake_up_process(dev->output.thread);
}
//*/
/*
static int g31_usb_output_stream_thread(void *param)
{
 int index,i,j;
 struct g31ddc_usb *dev;
 struct g31_io_stream *io_stream;
 struct g31_usb_buffer *usb_buffer;
 struct g31_usb_buffers *usb_buffers;
 unsigned long flags;
 int usb_buffer_size;
 int no_data;
 int status;
 uint32_t divider;
 struct g31_usb_output_packet *packet;
 uint8_t dac,dac2;
 uint8_t data;
 uint32_t bytes_to_copy;
 struct g31_io_buffer **io_buffers;
 uint32_t packet_size;
 uint32_t size[G31_IO_STREAM_OUT_COUNT];
 uint8_t *mem[G31_IO_STREAM_OUT_COUNT];
 uint8_t enabled[G31_IO_STREAM_OUT_COUNT];
 uint8_t first_packet[G31_IO_STREAM_OUT_COUNT];
 uint32_t ddc1_divider[G31_CHANNEL_COUNT];


    dev=param;

    usb_buffer=NULL;
    usb_buffers=&dev->output.avail_usb_buffers;
    index=0;
    usb_buffer_size=0;
    dac=0;
    packet=0;

    for(i=0;i<G31_IO_STREAM_OUT_COUNT;i++)
    {
        first_packet[i]=1;
        enabled[i]=0;
        size[i]=0;
        mem[i]=NULL;
    }

    for(i=0;i<G31_CHANNEL_COUNT;i++)
    {
        ddc1_divider[i]=0;
    }

    io_buffers=dev->io_buffer+G31_IO_STREAM_OUT_FIRST;

    while(!kthread_should_stop())
    {
        if(usb_buffer==NULL)
        {
            spin_lock_irqsave(&usb_buffers->spinlock,flags);

            if(unlikely(list_empty(&usb_buffers->list)))
            {
                spin_unlock_irqrestore(&usb_buffers->spinlock,flags);

                set_current_state(TASK_INTERRUPTIBLE);

                schedule();

                set_current_state(TASK_RUNNING);

                continue;
            }
            else
            {
                usb_buffer=list_entry(usb_buffers->list.next,struct g31_usb_buffer,list_entry);
                list_del(&usb_buffer->list_entry);
                spin_unlock_irqrestore(&usb_buffers->spinlock,flags);

                usb_buffer_size=0;

                packet=(struct g31_usb_output_packet*)usb_buffer->mem;
            }
        }

        data=1;

        for(i=0;i<G31_CHANNEL_COUNT;i++)
        {
            divider=dev->output.ddc1_divider[i];

            if(ddc1_divider[i]!=divider)
            {
                if((ddc1_divider[i] & 0xFFFF)!=0 && divider!=0)
                {
                    packet->sync=cpu_to_le16(0x4041);
                    packet->command=cpu_to_le16(1);
                    packet->address=cpu_to_le16(USB_REG_DDC2_INPUT_SWITCH_0+(USB_REG_DDC2_INPUT_SWITCH_1-USB_REG_DDC2_INPUT_SWITCH_0)*i);
                    packet->length=cpu_to_le16(3+1+2);
                    packet->data32[0]=cpu_to_le32(ddc1_divider[i] & 0xFFFF0000);

                    usb_buffer_size+=2*(3+1+2);

                    packet=(struct g31_usb_output_packet*)((uintptr_t)packet+2*(3+1+2));
                }

                first_packet[i]=1;

                ddc1_divider[i]=divider;

                packet->sync=cpu_to_le16(0x4041);
                packet->command=cpu_to_le16(1);
                packet->address=cpu_to_le16(USB_REG_DDC2_INPUT_SWITCH_0+(USB_REG_DDC2_INPUT_SWITCH_1-USB_REG_DDC2_INPUT_SWITCH_0)*i);
                packet->length=cpu_to_le16(3+1+2);
                packet->data32[0]=cpu_to_le32(divider);

                usb_buffer_size+=2*(3+1+2);

                packet=(struct g31_usb_output_packet*)((uintptr_t)packet+2*(3+1+2));
                enabled[i]=(divider & 0xFFFF)?1:0;
                data=0;
            }
        }

        dac2=dev->g31_dev.dac;

        if(dac!=dac2)
        {
            dac=dac2;

            packet->sync=cpu_to_le16(0x4041);
            packet->command=cpu_to_le16(1);
            packet->address=cpu_to_le16(USB_REG_DAC_CONTROL);
            packet->length=cpu_to_le16(3+1+2);
            packet->data32[0]=cpu_to_le32(dac?dev->output.dac_control:0);

            packet=(struct g31_usb_output_packet*)((uintptr_t)packet+2*(3+1+2));
            enabled[G31_IO_STREAM_OUT_DAC-G31_IO_STREAM_OUT_FIRST]=dac;
            data=0;
        }

        if(likely(data))
        {
            for(i=0;i<10;i++)
            {
                no_data=1;

                for(j=0;j<G31_IO_STREAM_OUT_COUNT;j++)
                {
                    if(enabled[index])
                    {
                        io_stream=&dev->g31_dev.io_stream[G31_IO_STREAM_OUT_FIRST+index];

                        down_read(&io_stream->rw_semaphore);

                        if(io_stream->active)
                        {
                            bytes_to_copy=dev->output.bytes_per_packet[index];

                            if(likely((usb_buffer_size+bytes_to_copy)<=G31_USB_OUTPUT_BUFFER_SIZE))
                            {
                                packet_size=0;

                                while(bytes_to_copy!=0)
                                {
                                    if(io_buffers[index]==NULL)
                                    {
                                        io_buffers[index]=g31_io_stream_get_pending_buffer(io_stream);

                                        if(unlikely(io_buffers[index]==NULL))
                                        {
                                            break;
                                        }

                                        if(unlikely(g31_io_buffer_map(io_buffers[index])<0))
                                        {
                                            g31_io_buffer_free(io_buffers[index]);
                                            io_buffers[index]=NULL;
                                            break;
                                        }

                                        size[index]=io_buffers[index]->size;
                                        mem[index]=io_buffers[index]->mem;
                                    }


                                    if(bytes_to_copy<size[index])
                                    {
                                        memcpy(packet->data8+packet_size,mem[index],bytes_to_copy);
                                        packet_size+=bytes_to_copy;
                                        size[index]-=bytes_to_copy;
                                        mem[index]+=bytes_to_copy;
                                        break;
                                    }

                                    memcpy(packet->data8+packet_size,mem[index],size[index]);
                                    packet_size+=size[index];
                                    bytes_to_copy-=size[index];

                                    g31_io_stream_put_complete_buffer(io_stream,io_buffers[index]);
                                    io_buffers[index]=NULL;
                                }

                                packet_size&=~1;

                                if(packet_size!=0)
                                {
                                    packet_size+=2*(3+1);

                                    packet->sync=cpu_to_le16(0x4041);
                                    packet->length=cpu_to_le16(packet_size>>1);
                                    packet->address=cpu_to_le16(index);

                                    if(first_packet[index])
                                    {
                                        packet->command=cpu_to_le16(3);
                                        first_packet[index]=0;
                                    }
                                    else
                                    {
                                        packet->command=0;
                                    }

                                    usb_buffer_size+=packet_size;
                                    packet=(struct g31_usb_output_packet*)((uintptr_t)packet+packet_size);

                                    no_data=0;
                                }
                            }
                            else
                            {
                                up_read(&io_stream->rw_semaphore);
                                no_data=1;
                                break;
                            }
                        }

                        up_read(&io_stream->rw_semaphore);
                    }

                    index=(index+1)%G31_IO_STREAM_OUT_COUNT;
                }

                if(no_data)
                {
                    break;
                }
            }
        }

        if(usb_buffer_size!=0)
        {
            usb_fill_bulk_urb(usb_buffer->urb,dev->usb_dev,
                              usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),
                              usb_buffer->mem,
                              usb_buffer_size,g31_usb_out_urb_complete_callback,
                              usb_buffer);

            usb_buffer->pending=1;

            status=usb_submit_urb(usb_buffer->urb,GFP_KERNEL);

            if(unlikely(status<0))
            {
                usb_buffer->pending=0;
                break;
            }

            usb_buffer=NULL;
        }
        else
        {
            set_current_state(TASK_INTERRUPTIBLE);

            schedule();

            set_current_state(TASK_RUNNING);
        }
    }

    while(usb_buffer==NULL)
    {
        spin_lock_irqsave(&usb_buffers->spinlock,flags);

        if(list_empty(&usb_buffers->list))
        {
            spin_unlock_irqrestore(&usb_buffers->spinlock,flags);

            set_current_state(TASK_INTERRUPTIBLE);

            schedule();

            set_current_state(TASK_RUNNING);
        }
        else
        {
            usb_buffer=list_entry(usb_buffers->list.next,struct g31_usb_buffer,list_entry);
            list_del(&usb_buffer->list_entry);

            spin_unlock_irqrestore(&usb_buffers->spinlock,flags);
        }
    }

    usb_buffer_size=0;
    packet=(struct g31_usb_output_packet*)usb_buffer->mem;

    for(i=0;i<G31_CHANNEL_COUNT;i++)
    {
        if(enabled[i])
        {
            packet->sync=cpu_to_le16(0x4041);
            packet->length=cpu_to_le16(3+1+2);
            packet->command=cpu_to_le16(1);
            packet->address=cpu_to_le16(USB_REG_DDC2_INPUT_SWITCH_0+(USB_REG_DDC2_INPUT_SWITCH_1-USB_REG_DDC2_INPUT_SWITCH_0)*i);
            packet->data32[0]=cpu_to_le32(ddc1_divider[i] & 0xFFFF0000);

            usb_buffer_size+=2*(3+1+2);

            packet=(struct g31_usb_output_packet*)((uintptr_t)packet+2*(3+1+2));
        }
    }

    if(enabled[i]) //DAC
    {
        packet->sync=cpu_to_le16(0x4041);
        packet->length=cpu_to_le16(3+1+2);
        packet->command=cpu_to_le16(1);
        packet->address=cpu_to_le16(USB_REG_DAC_CONTROL);
        packet->data32[0]=0;

        usb_buffer_size+=2*(3+1+2);
    }

    if(usb_buffer_size!=0)
    {
        usb_fill_bulk_urb(usb_buffer->urb,dev->usb_dev,
                          usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_FPGA),
                          usb_buffer->mem,
                          usb_buffer_size,g31_usb_out_urb_complete_callback,
                          usb_buffer);

        usb_buffer->pending=1;

        status=usb_submit_urb(usb_buffer->urb,GFP_KERNEL);

        if(unlikely(status<0))
        {
            usb_buffer->pending=0;
        }
    }

    for(i=0;i<G31_USB_OUTPUT_BUFFER_COUNT;i++)
    {
        while(dev->output.buffer[i].pending)
        {
            set_current_state(TASK_INTERRUPTIBLE);

            schedule();

            set_current_state(TASK_RUNNING);
        }
    }

    while(!kthread_should_stop())
    {
        set_current_state(TASK_INTERRUPTIBLE);

        schedule();

        set_current_state(TASK_RUNNING);
    }

    return 0;
}
//*/

static int g31_usb_start_stream(struct g31ddc_usb *dev,uint32_t stream_index)
{
 struct task_struct *thread;
 int i;
 int buffer_size;

    if(G31_IS_INPUT_STREAM(stream_index))
    {
        if(dev->input.stream_counter==0)
        {
            buffer_size=G31_USB_INPUT_BUFFER_MAX_SIZE;

            for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
            {
                dev->input.buffer[i].pending=0;

                dev->input.buffer[i].urb=usb_alloc_urb(0,GFP_KERNEL);

                if(unlikely(dev->input.buffer[i].urb==NULL))
                {
                    while(i--)
                    {
                        kfree(dev->input.buffer[i].mem);
                        dev->input.buffer[i].mem=NULL;

                        usb_free_urb(dev->input.buffer[i].urb);
                        dev->input.buffer[i].urb=NULL;
                    }

                    return -ENOMEM;
                }

                dev->input.buffer[i].mem=kmalloc(buffer_size+sizeof(struct g31_usb_input_packet),GFP_KERNEL);

                if(dev->input.buffer[i].mem==NULL)
                {
                    usb_free_urb(dev->input.buffer[i].urb);
                    dev->input.buffer[i].urb=NULL;

                    while(i--)
                    {
                        kfree(dev->input.buffer[i].mem);
                        dev->input.buffer[i].mem=NULL;

                        usb_free_urb(dev->input.buffer[i].urb);
                        dev->input.buffer[i].urb=NULL;
                    }

                    return -ENOMEM;
                }
            }

            INIT_LIST_HEAD(&dev->input.complete_usb_buffers.list);

            thread=kthread_create(g31_usb_input_stream_thread,dev,G31_DEVICE_NAME);

            if(unlikely(IS_ERR(thread)))
            {
                for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
                {
                    kfree(dev->input.buffer[i].mem);
                    dev->input.buffer[i].mem=NULL;

                    usb_free_urb(dev->input.buffer[i].urb);
                    dev->input.buffer[i].urb=NULL;
                }

                return PTR_ERR(thread);
            }

            dev->input.thread=thread;

            wake_up_process(thread);
        }

        dev->input.stream_counter++;

        dev->g31_dev.io_stream[stream_index].active=1;

        if(stream_index>=G31_IO_STREAM_IN_FIRST && stream_index<=G31_IO_STREAM_IN_LAST)
        {
            g31_usb_update_input_buffer_size(dev);
        }
    }
 /*
    else
    {
        bytes_per_packet=((dev->g31_dev.io_stream[stream_index].bytes_per_sec/1000+11)/12)*12;

        if(unlikely(bytes_per_packet>(G31_USB_OUTPUT_BUFFER_SIZE-sizeof(struct g31_usb_output_packet))))
        {
            bytes_per_packet=G31_USB_OUTPUT_BUFFER_SIZE-sizeof(struct g31_usb_output_packet);
        }
        else if(unlikely(bytes_per_packet<168))
        {
            bytes_per_packet=168;
        }

        dev->output.bytes_per_packet[stream_index-G31_IO_STREAM_OUT_FIRST]=bytes_per_packet;

        if(dev->output.stream_counter==0)
        {
            INIT_LIST_HEAD(&dev->output.avail_usb_buffers.list);

            for(i=0;i<G31_USB_OUTPUT_BUFFER_COUNT;i++)
            {
                dev->output.buffer[i].pending=0;

                dev->output.buffer[i].urb=usb_alloc_urb(0,GFP_KERNEL);

                if(unlikely(dev->output.buffer[i].urb==NULL))
                {
                    while(i--)
                    {
                        kfree(dev->output.buffer[i].mem);
                        dev->output.buffer[i].mem=NULL;

                        usb_free_urb(dev->output.buffer[i].urb);
                        dev->output.buffer[i].urb=NULL;
                    }

                    return -ENOMEM;
                }

                dev->output.buffer[i].mem=kmalloc(G31_USB_OUTPUT_BUFFER_SIZE,GFP_KERNEL);

                if(dev->output.buffer[i].mem==NULL)
                {
                    usb_free_urb(dev->output.buffer[i].urb);
                    dev->output.buffer[i].urb=NULL;

                    while(i--)
                    {
                        kfree(dev->output.buffer[i].mem);
                        dev->output.buffer[i].mem=NULL;

                        usb_free_urb(dev->output.buffer[i].urb);
                        dev->output.buffer[i].urb=NULL;
                    }

                    return -ENOMEM;
                }

                list_add_tail(&dev->output.buffer[i].list_entry,&dev->output.avail_usb_buffers.list);
            }


            thread=kthread_create(g31_usb_output_stream_thread,dev,G31_DEVICE_NAME);

            if(unlikely(IS_ERR(thread)))
            {
                for(i=0;i<G31_USB_OUTPUT_BUFFER_COUNT;i++)
                {
                    kfree(dev->output.buffer[i].mem);
                    dev->output.buffer[i].mem=NULL;

                    usb_free_urb(dev->output.buffer[i].urb);
                    dev->output.buffer[i].urb=NULL;
                }

                return PTR_ERR(thread);
            }

            dev->output.thread=thread;

            wake_up_process(thread);
        }

        dev->output.stream_counter++;

        dev->g31_dev.io_stream[stream_index].active=1;
    }
//*/
    return 0;
}

static void g31_usb_stop_stream(struct g31ddc_usb *dev,uint32_t stream_index)
{
 struct g31_io_stream *io_stream;
 int i;

    io_stream=dev->g31_dev.io_stream+stream_index;

    down_write(&io_stream->rw_semaphore);

    io_stream->active=0;

    up(&io_stream->complete_buffers.semaphore);

    up_write(&io_stream->rw_semaphore);

    if(G31_IS_INPUT_STREAM(stream_index))
    {
        dev->input.stream_counter--;

        if(dev->input.stream_counter==0)
        {
            kthread_stop(dev->input.thread);

            dev->input.thread=NULL;

            for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
            {
                kfree(dev->input.buffer[i].mem);
                dev->input.buffer[i].mem=NULL;

                usb_free_urb(dev->input.buffer[i].urb);
                dev->input.buffer[i].urb=NULL;
            }
        }

        g31_io_stream_free_buffers(io_stream);

        if(dev->io_buffer[stream_index]!=NULL)
        {
            g31_io_buffer_free(dev->io_buffer[stream_index]);
            dev->io_buffer[stream_index]=NULL;
        }

        g31_usb_update_input_buffer_size(dev);
    }
    else
    {
        dev->output.stream_counter--;

        if(dev->output.stream_counter==0)
        {
            kthread_stop(dev->output.thread);

            dev->output.thread=NULL;

            for(i=0;i<G31_USB_OUTPUT_BUFFER_COUNT;i++)
            {
                kfree(dev->output.buffer[i].mem);
                dev->output.buffer[i].mem=NULL;

                usb_free_urb(dev->output.buffer[i].urb);
                dev->output.buffer[i].urb=NULL;
            }
        }

        g31_io_stream_free_buffers(io_stream);

        if(dev->io_buffer[stream_index]!=NULL)
        {
            g31_io_buffer_free(dev->io_buffer[stream_index]);
            dev->io_buffer[stream_index]=NULL;
        }
    }
}

static int g31_usb_if_start(struct g31ddc_usb *dev,uint16_t period)
{
int status;
uint16_t Packet[2];

Packet[0]=1;
Packet[1]=period;

dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].buffer_size=65536*sizeof(int16_t)+sizeof(uint16_t);
dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].bytes_per_sec=dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].buffer_size*1000/period;
status=g31_usb_fpga_cmd(dev,4,DeviceUSB_GetNextIFCommandId(dev),Packet,sizeof(Packet));

/*
 uint16_t addresses[2];
 uint32_t values[2];

    dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].buffer_size=65536*sizeof(int16_t)+sizeof(uint16_t);
    dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].bytes_per_sec=dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].buffer_size*1000/period;

    addresses[0]=USB_REG_COMMAND_ID;
    values[0]=++dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].id;

    addresses[1]=USB_REG_IF_BUFF_CTRL;
    values[1]=period;

    status=g31_usb_set_regs(dev,addresses,values,2);

    if(unlikely(status<0))
    {
        return status;
    }
//*/

status=g31_usb_start_stream(dev,G31_IO_STREAM_IN_IF);
if(unlikely(status<0)) {
	Packet[0]=0;
	Packet[1]=0;
	status=g31_usb_fpga_cmd(dev,4,DeviceUSB_GetNextIFCommandId(dev),Packet,sizeof(Packet));
	}
return status;
}

static void g31_usb_if_stop(struct g31ddc_usb *dev)
{
uint16_t Packet[2];
int status;

if(likely(dev->g31_dev.io_stream[G31_IO_STREAM_IN_IF].active)) {
	g31_usb_stop_stream(dev,G31_IO_STREAM_IN_IF);
	Packet[0]=1;
	Packet[1]=0;
	status=g31_usb_fpga_cmd(dev,4,DeviceUSB_GetNextIFCommandId(dev),Packet,sizeof(Packet));
	}
}

static int g31_usb_put_buffer(struct g31ddc_usb *dev,uint32_t stream_index,struct g31_io_buffer *io_buffer)
{
if(likely(G31_IS_INPUT_STREAM(stream_index))) g31_io_stream_put_pending_buffer(&dev->g31_dev.io_stream[stream_index],io_buffer);
else {
	if(g31_io_stream_put_pending_buffer2(&dev->g31_dev.io_stream[stream_index],io_buffer)) wake_up_process(dev->output.thread);
    }

return 0;
}

static int g31_usb_mmap(struct g31ddc_usb *dev,struct vm_area_struct *vma)
{
return -ENOSYS;
}

static int g31_usb_get_adc_level(struct g31ddc_usb *dev,uint16_t *level)
{
*level=dev->ADC_Level;
return 0;
}
static int g31_usb_get_adc_sample_rate(struct g31ddc_usb *dev,uint32_t *sr)
{
*sr=dev->TCXO_Freq;
return 0;
}

static int g31_usb_wd(struct g31ddc_usb *dev)
{
 uint8_t state[2];
 int status;

    status=usb_control_msg(dev->usb_dev, usb_rcvctrlpipe(dev->usb_dev,0),0x10,USB_TYPE_VENDOR|USB_DIR_IN,0,0,state,sizeof(state),G31_USB_TIMEOUT);

    if(unlikely(status<0))
    {
        return -1;
    }

    return state[0]>=68;
}

static int g31_usb_ddc1_setup(struct g31ddc_usb *dev,uint32_t channel,int enabled,uint32_t *coeffs)
{
struct g31_ddc1_params *params;
int status,CoeffIndex;
uint32_t i,j,tmp;
uint16_t CmdId,Data[55];
uint16_t Coeffs[1024],MSB,LSB;

params=dev->g31_dev.ddc1_params+channel;

if(coeffs!=NULL) {
	Data[0]=params->control_word & 0xFFFE;
	Data[1]=(uint16_t)(params->tuning_word>>16);      // freqH
	Data[2]=(uint16_t)(params->tuning_word&0xFFFF);   // FreqL
	Data[3]=0;
	Data[4]=0;
	Data[5]=params->cic1;
	Data[6]=params->cic2;
	Data[7]=params->fir_len;//FIRParams->FIRLength;// DDC_Registers_Low->FIRlength;
	Data[8]=params->fir_control;//FIRParams->FIRControl;// DDC_Registers_Low->FIRcontrol;
	CmdId=DeviceUSB_GetNextIFCommandId(dev);
	status=g31_usb_fpga_cmd(dev,6,CmdId,Data,9*2);

	if(unlikely(status)) return status;

	for(i=params->fir_len>>1;i!=0;i--) {
		tmp=coeffs[i-1];//DDC_Coeffs_Low[i-1];
		LSB=(tmp&3)<<14;
		MSB=(tmp>>2);
		j=params->fir_len-(i<<1);
		Coeffs[j]=LSB;
		Coeffs[j+1]=MSB;
		}
	CoeffIndex=(int)(params->fir_len%54)-54;    
	Data[0]=0;
	j=1;

	while(CoeffIndex<=(int32_t)params->fir_len) {
		if(j==55) {
			CmdId=DeviceUSB_GetNextIFCommandId(dev);
			status=g31_usb_fpga_cmd(dev,7,CmdId,Data,sizeof(Data));

			if (status) return status;
			if (CoeffIndex==(int32_t)params->fir_len) break;

			Data[0]=0;
			j=1;
	        }

		if(CoeffIndex>=0) Data[j++]=Coeffs[CoeffIndex];
		else Data[j++]=0;

		CoeffIndex++;
		}
	}

if (enabled) Data[0]=params->control_word;
else Data[0]=params->control_word& 0xFFFE;
Data[1]=(uint16_t)(params->tuning_word>>16);      // freqH
Data[2]=(uint16_t)(params->tuning_word&0xFFFF);   // FreqL
Data[3]=0;
Data[4]=0;
Data[5]=params->cic1;
Data[6]=params->cic2;
Data[7]=params->fir_len;
Data[8]=params->fir_control;
CmdId=DeviceUSB_GetNextIFCommandId(dev);

return g31_usb_fpga_cmd(dev,6,CmdId,Data,9*2);

/*
        status=g31_usb_set_reg(dev,USB_REG_COEFF_WR_CTRL,1<<channel);

        for(i=0,j=(params->fir_len>>1)-1;i<params->fir_len>>2;i++,j--)
        {
            tmp=coeffs[i];
            coeffs[i]=coeffs[j];
            coeffs[j]=tmp;
        }

        status=g31_usb_set_reg_rep(dev,USB_REG_DDC1_FIR_COEFF_0+channel,coeffs,params->fir_len>>1);

        if(unlikely(status)) return status;

        if(enabled) status=g31_usb_set_reg(dev,USB_REG_DDC1_CTRL_0+reg_offset,params->control_word);
        else status=g31_usb_set_reg(dev,USB_REG_DDC1_CTRL_0+reg_offset,params->control_word & 0xFFFE);
    }
    else {
        if(enabled) values[0]=params->control_word;
        status=g31_usb_set_regs(dev,addresses,values,4);
    }
    return status;
//*/
}

static int g31_usb_ddc1_start(struct g31ddc_usb *dev,uint32_t channel,uint32_t *coeffs)
{
int status;

status=g31_usb_ddc1_setup(dev,channel,1,coeffs);
if(unlikely(status<0)) return status;
status=g31_usb_start_stream(dev,G31_IO_STREAM_IN_DDC+channel);
if(unlikely(status<0)) g31_usb_ddc1_setup(dev,channel,0,NULL);
return status;;
}

static void g31_usb_ddc1_stop(struct g31ddc_usb *dev,uint32_t channel)
{
if (likely(dev->g31_dev.io_stream[G31_IO_STREAM_IN_DDC+channel].active)) {
	g31_usb_ddc1_setup(dev,channel,0,NULL);
	g31_usb_stop_stream(dev,G31_IO_STREAM_IN_DDC+channel);
	}
}

static int g31_usb_ddc1_update_params(struct g31ddc_usb *dev,uint32_t channel)
{
return g31_usb_ddc1_setup(dev,channel,1,NULL);
}

#define CHUNKSZ 1024
static int g31_usb_read_config(struct g31ddc_usb *dev)
{
 uint32_t size;
 uint32_t to_read;
 uint16_t offset;
 uint8_t *p;
 int status;

    size=8192;
    p=dev->g31_dev.config+4;
    offset=0;

    while(size>1)
    {
        if(size>=CHUNKSZ)
        {
            to_read=CHUNKSZ;
        }
        else
        {
            to_read=size;
        }

        status=usb_control_msg(dev->usb_dev,usb_rcvctrlpipe(dev->usb_dev,0),0x0D,USB_TYPE_VENDOR|USB_DIR_IN,offset,0,p,to_read,G31_USB_TIMEOUT);

        if(unlikely(status<=0))
        {
			G31_ERROR("read config data error at %d, status=%d\n",(int)offset,status);
            return -EIO;
        }
//        else G31_ERROR("read config data OK at %d, status=%d\n",(int)offset,status);

        p+=status;
        size-=status;
        offset+=status;
    }

    return 0;
}

static void g31_usb_cancel_buffers(struct g31ddc_usb *dev,int stream_index)
{
 struct g31_io_stream *io_stream;
 struct g31_io_buffer *io_buffer;

    io_stream=&dev->g31_dev.io_stream[stream_index];

    while((io_buffer=g31_io_stream_get_pending_buffer(io_stream)))
    {
        g31_io_stream_put_complete_buffer(io_stream,io_buffer);
    }
}

static int g31_usb_initialize(struct g31ddc_usb *dev)
{
 int i;
 int status;

    dev->g31_dev.func_delete=(g31_func_delete_t)g31_usb_delete;
    dev->g31_dev.func_open=(g31_func_open_t)g31_usb_open;
    dev->g31_dev.func_close=(g31_func_close_t)g31_usb_close;
    dev->g31_dev.func_set_power=(g31_func_set_power_t)g31_usb_set_power;
    dev->g31_dev.func_configure_fpga=(g31_func_configure_fpga_t)g31_usb_configure_fpga;
//    dev->g31_dev.func_set_spi_regs=(g31_func_set_spi_regs_t)g31_usb_set_spi_regs;
//    dev->g31_dev.func_set_spi2_regs=(g31_func_set_spi2_regs_t)g31_usb_set_spi2_regs;
    dev->g31_dev.func_rf_control1=(g31_func_rf_control)g31_usb_rf_control1;
    dev->g31_dev.func_rf_control2=(g31_func_rf_control2)g31_usb_rf_control2;
    dev->g31_dev.func_set_noise_blanker=(g31_func_set_noise_blanker_t)g31_usb_set_noise_blanker;
    dev->g31_dev.func_put_buffer=(g31_func_put_buffer_t)g31_usb_put_buffer;
    dev->g31_dev.func_if_start=(g31_func_if_start_t)g31_usb_if_start;
    dev->g31_dev.func_if_stop=(g31_func_if_stop_t)g31_usb_if_stop;
    dev->g31_dev.func_get_adc_level=(g31_func_get_adc_level_t)g31_usb_get_adc_level;
    dev->g31_dev.func_get_adc_sample_rate=(g31_func_get_adc_sample_rate_t)g31_usb_get_adc_sample_rate;
    dev->g31_dev.func_usb_control1_read=(g31_func_usb_control1_t)g31_usb_control1_read;
    dev->g31_dev.func_usb_control1_write=(g31_func_usb_control1_t)g31_usb_control1_write;
    dev->g31_dev.func_usb_control2_read=(g31_func_usb_control2_t)g31_usb_control2_read;
    dev->g31_dev.func_usb_control2_write=(g31_func_usb_control2_t)g31_usb_control2_write;
    dev->g31_dev.func_mmap=(g31_func_mmap_t)g31_usb_mmap;
    dev->g31_dev.func_wd=(g31_func_wd_t)g31_usb_wd;
    dev->g31_dev.func_ddc1_start=(g31_func_ddc1_start_t)g31_usb_ddc1_start;
    dev->g31_dev.func_ddc1_stop=(g31_func_ddc1_stop_t)g31_usb_ddc1_stop;
    dev->g31_dev.func_ddc1_update_params=(g31_func_ddc1_update_params_t)g31_usb_ddc1_update_params;
    dev->g31_dev.func_cancel_buffers=(g31_func_cancel_buffers_t)g31_usb_cancel_buffers;

    spin_lock_init(&dev->input.complete_usb_buffers.spinlock);

	dev->IFCommandId=0;

    dev->input.buffer_size=G31_USB_INPUT_BUFFER_MAX_SIZE;

    for(i=0;i<G31_USB_INPUT_BUFFER_COUNT;i++)
    {
        dev->input.buffer[i].dev=dev;
    }

    spin_lock_init(&dev->output.avail_usb_buffers.spinlock);

    for(i=0;i<G31_USB_OUTPUT_BUFFER_COUNT;i++)
    {
        dev->output.buffer[i].dev=dev;
    }

    status=g31_usb_setup(dev);

    if(unlikely(status))
    {
        return status;
    }

    status=g31_usb_read_config(dev);

    if(unlikely(status))
    {
		int len;
		uint32_t Packet[2]={0x0,0xB}; 
		uint8_t RcvData[16]={0};
    
		//*
		status=usb_bulk_msg(dev->usb_dev,usb_sndbulkpipe(dev->usb_dev,G31_USB_EP_OUT_CTRL),Packet,8,&len,G31_USB_TIMEOUT); 
		if (status<0) G31_ERROR("SERNUM send error %d\n",status);
		else G31_ERROR("SERNUM retr send OK, status=%d, len=%d\n",status,len);

		if (!status) {
			status=usb_bulk_msg(dev->usb_dev,usb_rcvbulkpipe(dev->usb_dev,G31_USB_EP_IN_CTRL),RcvData,16,&len,G31_USB_TIMEOUT); 
			if (status<0) G31_ERROR("SERNUM rcv error %d\n",status);
			else G31_ERROR("SERNUM rcv OK, status=%d, len=%d\n",status,len);
			
			if (!status) {
				memset(dev->g31_dev.config+4,0,8192);
				memcpy(dev->g31_dev.config+12,RcvData+8,8);
				dev->g31_dev.config[5]=6;	// receiver model
			
				G31_ERROR("Device config is created\n");
				}			
			}
		//*/
        
        return status;
    }

    return 0;
}

static int g31_usb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
 int status;
 struct g31ddc_usb *dev;
 struct usb_device *usb_dev;

    usb_dev=interface_to_usbdev(interface);

    if(unlikely(id->idVendor!=G31_USB_VID || id->idProduct!=G31_USB_PID))
    {
        G31_ERROR("unsuported device!\n");
        return -ENODEV;
    }

    if(unlikely(usb_dev->speed<USB_SPEED_HIGH))
    {
        G31_ERROR("device is not running at top speed!\n");
        return -ENODEV;
    }

    dev=kmalloc(sizeof(*dev),GFP_KERNEL);

    if(unlikely(dev==NULL))
    {
        G31_ERROR("failed to allocate memory for device object!\n");
        return -ENOMEM;
    }

    memset(dev,0,sizeof(*dev));

    dev->g31_dev.config[0]=G31_INTERFACE_TYPE_USB;
    dev->usb_dev=usb_get_dev(usb_dev);
    dev->usb_interface=usb_get_intf(interface);

    usb_set_intfdata(interface,dev);

    g31_initialize(&dev->g31_dev);

    status=g31_usb_initialize(dev);

    if(unlikely(status<0))
    {
        g31_uninitialize(&dev->g31_dev);
        usb_set_intfdata(interface,NULL);
        usb_put_intf(dev->usb_interface);
        usb_put_dev(dev->usb_dev);
        kfree(dev);
        return status;
    }

    status=g31_register(&dev->g31_dev);

    if(unlikely(status<0))
    {
        g31_delete(&dev->g31_dev);
        return status;
    }

    return 0;
}

static void g31_usb_disconnect(struct usb_interface *interface)
{
 struct g31ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    usb_set_intfdata(interface,NULL);

    if(likely(dev!=NULL))
    {
        dev->g31_dev.state.connected=0;

        wake_up_all(&dev->g31_dev.power_wait_queue);

        g31_unregister(&dev->g31_dev);
    }
}

#ifdef CONFIG_PM

static int g31_usb_suspend(struct usb_interface *interface,pm_message_t message)
{
 struct g31ddc_usb *dev;

G31_ERROR("Suspending...\n");

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        G31_DEV_LOCK2;

        if(dev->g31_dev.state.power)
        {
            dev->g31_dev.state.suspended=1;
            g31_set_power(&dev->g31_dev,0,0);
        }

        G31_DEV_UNLOCK2;
    }

    return 0;
}

static int g31_usb_resume(struct usb_interface *interface)
{
 struct g31ddc_usb *dev;
 int status;

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        status=g31_usb_setup(dev);

        if(unlikely(status!=0))
        {
            return status;
        }
    }

    return 0;
}

#endif //CONFIG_PM


static struct usb_device_id g31_usb_id_table[]=
{
    {USB_DEVICE(G31_USB_VID,G31_USB_PID)},
    {},
};

static struct usb_driver g31_usb_driver=
{
    .name=G31_DEVICE_NAME,
    .probe=g31_usb_probe,
    .disconnect=g31_usb_disconnect,
    .id_table=g31_usb_id_table,
#ifdef CONFIG_PM
    .suspend=g31_usb_suspend,
    .resume=g31_usb_resume,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
    .reset_resume=g31_usb_resume
#endif //KERNEL_VERSION(2,6,23)
#endif //CONFIG_PM
};

MODULE_DEVICE_TABLE(usb,g31_usb_id_table);

#endif //__G31_USB__

static int g31_fo_open(struct inode *node, struct file *file)
{
 uint32_t index;
 int status;
 struct g31ddc *dev;

    index=iminor(file->f_path.dentry->d_inode);

    G31_REG_LOCK;

    dev=g31_registered_devices[index];

    if(unlikely(dev==NULL))
    {
        G31_REG_UNLOCK;
        return -ENXIO;
    }

    get_device(&dev->dev);

    G31_DEV_LOCK;

    if(unlikely(dev->state.open))
    {
        G31_DEV_UNLOCK;
        G31_REG_UNLOCK;
        put_device(&dev->dev);
        return -EBUSY;
    }

    dev->state.open=1;

    G31_DEV_UNLOCK;

    dev->workqueue=create_singlethread_workqueue(G31_DEVICE_NAME);

    if(unlikely(dev->workqueue==NULL))
    {
        dev->state.open=0;
        G31_REG_UNLOCK;
        put_device(&dev->dev);
        return -EIO;
    }

    G31_DEV_LOCK;

    status=dev->func_open(dev);

    if(unlikely(status<0))
    {
        dev->state.open=0;
    }

    G31_DEV_UNLOCK;

    if(likely(status>=0))
    {
        file->private_data=dev;

        dev->wd_counter=0;

        dev->wd_timer.expires=jiffies+2*HZ;
        add_timer(&dev->wd_timer);
    }
    else
    {
        destroy_workqueue(dev->workqueue);
        dev->workqueue=NULL;
    }

    G31_REG_UNLOCK;

    if(unlikely(status<0))
    {
        put_device(&dev->dev);
    }

    return status;
}

static int g31_fo_release(struct inode *node, struct file *file)
{
 struct g31ddc *dev;

    dev=file->private_data;

    if(likely(dev!=NULL))
    {
        G31_DEV_LOCK;

        g31_set_power(dev,0,0);

        dev->state.open=0;

        dev->func_close(dev);

        wake_up_all(&dev->power_wait_queue);

        G31_DEV_UNLOCK;

        del_timer_sync(&dev->wd_timer);

        if(likely(dev->workqueue!=NULL))
        {
            flush_workqueue(dev->workqueue);
            destroy_workqueue(dev->workqueue);
            dev->workqueue=NULL;
        }

        put_device(&dev->dev);

        file->private_data=NULL;
    }

    return 0;
}



static int __init g31_init(void)
{
 int status;

    g31_mutex_init(&g31_reg_mutex);

    printk(KERN_INFO G31_DESCRIPTION "\n");

    status=alloc_chrdev_region(&g31_dev0,0,G31_MAX_DEVICE_COUNT,G31_DEVICE_NAME);

    if(status<0)
    {
        G31_ERROR("unable to allocate chrdev region!\n");
        goto error_mutex_destroy;
    }

    status=class_register(&g31_class);

    if(status<0)
    {
        G31_ERROR("failed to register class!\n");
        goto error_unregister_chrdev_region;
    }

#ifdef __G31_USB__

    status=usb_register(&g31_usb_driver);

    if(unlikely(status<0))
    {
        G31_ERROR("failed to register usb driver!\n");
        goto error_class_unregister;
    }

#endif //__G31_USB__

#ifdef __G31_PCI__
    status=pci_register_driver(&g31_pci_driver);

    if(unlikely(status<0))
    {
        G31_ERROR("failed to register pci driver!\n");
        goto error_usb_deregister;
    }
#endif //__G31_PCI__

    return 0;

//    error_usb_deregister:
#ifdef __G31_USB__
        usb_deregister(&g31_usb_driver);

    error_class_unregister:
#endif
        class_unregister(&g31_class);

    error_unregister_chrdev_region:
        unregister_chrdev_region(g31_dev0,G31_MAX_DEVICE_COUNT);

    error_mutex_destroy:
        g31_mutex_destroy(&g31_reg_mutex);

    return status;
}

static void __exit g31_exit(void)
{
#ifdef __G31_PCI__
    pci_unregister_driver(&g31_pci_driver);
#endif //__G31_PCI__

#ifdef __G31_USB__
    usb_deregister(&g31_usb_driver);
#endif //__G31_USB__

    class_unregister(&g31_class);
    unregister_chrdev_region(g31_dev0,G31_MAX_DEVICE_COUNT);

    g31_mutex_destroy(&g31_reg_mutex);
}


static int g31_fo_mmap(struct file *file,struct vm_area_struct *vma)
{
 struct g31ddc *dev;

    dev=file->private_data;

    if(unlikely(dev==NULL || !dev->state.connected))
    {
        return -ENODEV;
    }

    return dev->func_mmap(dev,vma);
}

#ifdef HAVE_UNLOCKED_IOCTL
static long g31_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
static int g31_fo_ioctl(struct inode *node,struct file *file, unsigned int cmd, unsigned long arg)
#endif //HAVE_UNLOCKED_IOCTL
{
 int status;
 struct g31ddc *dev;
 uint8_t byte;
 uint16_t word;
 uint32_t dword;
 /*
 struct g31_spi2_registers spi2;
 struct g31_spi_registers spi;
 */
 struct g31_usb_control1 usb_control1;
 struct g31_usb_control2 usb_control2; 
 struct g31_rf_control rf;
 struct g31_rf_control2 rf2;
 struct g31_fpga_data fpga;
 struct g31_ddc1_start_params ddc1_start_params;
 //struct g31_dac_params dac_params;

    dev=file->private_data;

    if(unlikely(dev==NULL))
    {
        return -ENODEV;
    }

    switch(cmd)
    {
        case G31_IOCTL_DDC1_0_PUT_BUFFER:
            status=g31_put_buffer(dev,G31_IO_STREAM_IN_DDC,(void __user*)arg,1);
            break;

        case G31_IOCTL_IF_PUT_BUFFER:
            status=g31_put_buffer(dev,G31_IO_STREAM_IN_IF,(void __user*)arg,1);
            break;

        case G31_IOCTL_DDC1_0_GET_BUFFER:
            status=g31_get_complete_buffer(dev,G31_IO_STREAM_IN_DDC,arg);
            break;

        case G31_IOCTL_IF_GET_BUFFER:
            status=g31_get_complete_buffer(dev,G31_IO_STREAM_IN_IF,arg);
            break;

        case G31_IOCTL_GET_ADC_LEVEL:
            status=g31_get_adc_level(dev,&word);

            if(likely(status==0))
            {
                status=put_user(word,(uint16_t __user*)arg);
            }
            break;
        case G31_IOCTL_GET_ADC_SAMPLE_RATE:
            status=g31_get_adc_sample_rate(dev,&dword);

            if(likely(status==0))
            {
                status=put_user(dword,(uint32_t __user*)arg);
            }
            break;

        case G31_IOCTL_DDC1_0_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g31_ddc1_set_frequency(dev,0,dword);
            break;

        case G31_IOCTL_DDC1_0_START:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)arg,sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g31_ddc1_start(dev,0,&ddc1_start_params.params,(void __user*)ddc1_start_params.coeffs,
                    ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G31_IOCTL_DDC1_0_STOP:
            g31_ddc1_stop(dev,0);
            status=0;
            break;

        case G31_IOCTL_IF_START:
            if(unlikely(get_user(word,(uint16_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g31_if_start(dev,word);
            break;

        case G31_IOCTL_IF_STOP:
            g31_if_stop(dev);

            status=0;
            break;

        case G31_IOCTL_GET_STATE:
            G31_DEV_LOCK;

            status=put_user(dev->state,(struct g31_state __user*)arg);

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_SET_NOISE_BLANKER:
            if(unlikely(get_user(word,(uint16_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g31_set_noise_blanker(dev,word);
            break;

        case G31_IOCTL_IF_CANCEL_BUFFERS:
            g31_cancel_buffers(dev,G31_IO_STREAM_IN_IF);
            status=0;
            break;

        case G31_IOCTL_DDC1_0_CANCEL_BUFFERS:
            g31_cancel_buffers(dev,G31_IO_STREAM_IN_DDC);
            status=0;
            break;

        case G31_IOCTL_USB_CONTROL1_READ:
            if(unlikely(copy_from_user(&usb_control1,(struct g31_usb_control1 __user*)arg,sizeof(usb_control1))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control1_read(dev,usb_control1.params,(void __user*)usb_control1.buffer,usb_control1.size,usb_control1.transferred);

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL1_WRITE:
            if(unlikely(copy_from_user(&usb_control1,(struct g31_usb_control1 __user*)arg,sizeof(usb_control1))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control1_write(dev,usb_control1.params,(void __user*)usb_control1.buffer,usb_control1.size,usb_control1.transferred);

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL2_READ:
            if(unlikely(copy_from_user(&usb_control2,(struct g31_usb_control2 __user*)arg,sizeof(usb_control2))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control2_read(dev,(void __user*)usb_control2.buffer,usb_control2.size,usb_control2.transferred);

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL2_WRITE:
            if(unlikely(copy_from_user(&usb_control2,(struct g31_usb_control2 __user*)arg,sizeof(usb_control2))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control2_write(dev,(void __user*)usb_control2.buffer,usb_control2.size,usb_control2.transferred);

            G31_DEV_UNLOCK;
            break; 
            
        case G31_IOCTL_SET_POWER:
            if(unlikely(get_user(byte,(uint8_t __user*)arg)<0))
            {
                return -EFAULT;
            }

            status=g31_set_power(dev,byte,1);
            break;

        case G31_IOCTL_CONFIGURE_FPGA:
            if(unlikely(copy_from_user(&fpga,(struct g31_fpga_data __user*)arg,sizeof(fpga))!=0))
            {
                return -EFAULT;
            }

            status=g31_configure_fpga(dev,(void __user*)fpga.buffer,fpga.size);
            break;

        case G31_IOCTL_GET_CONFIG:
            status=copy_to_user((void __user*)arg,dev->config,sizeof(dev->config))?-EFAULT:0;
            break;

        case G31_IOCTL_RF_CONTROL1:
            if(unlikely(copy_from_user(&rf,(struct g31_rf_control __user*)arg,sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g31_rf_control1(dev,&rf);
            break;

        case G31_IOCTL_RF_CONTROL2:
            if(unlikely(copy_from_user(&rf2,(struct g31_rf_control2 __user*)arg,sizeof(rf2))!=0))
            {
                return -EFAULT;
            }

            status=g31_rf_control2(dev,&rf2);
            break;

        default:
            status=-EINVAL;
            break;
    }

    return status;
}

#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)

static long g31_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg)
{
 int status;
 struct g31ddc *dev;
 uint8_t byte;
 uint16_t word;
 uint32_t dword;
 /*
 struct g31_spi2_registers_compat spi2;
 struct g31_spi_registers_compat spi;
*/
 struct g31_rf_control rf;
 struct g31_rf_control2 rf2;
 struct g31_fpga_data_compat fpga;
 struct g31_ddc1_start_params_compat ddc1_start_params;
 //struct g31_dac_params dac_params;
 struct g31_usb_control1_compat usb_control1;
 struct g31_usb_control2_compat usb_control2; 
 
    dev=file->private_data;

    if(unlikely(dev==NULL))
    {
        return -ENODEV;
    }

    switch(cmd)
    {
        case G31_IOCTL_DDC1_0_PUT_BUFFER_COMPAT:
            status=g31_put_buffer(dev,G31_IO_STREAM_IN_DDC,compat_ptr(arg),1);
            break;

        case G31_IOCTL_IF_PUT_BUFFER_COMPAT:
            status=g31_put_buffer(dev,G31_IO_STREAM_IN_IF,compat_ptr(arg),1);
            break;

        case G31_IOCTL_DDC1_0_GET_BUFFER_COMPAT:
            status=g31_get_complete_buffer_compat(dev,G31_IO_STREAM_IN_DDC,arg);
            break;

        case G31_IOCTL_IF_GET_BUFFER_COMPAT:
            status=g31_get_complete_buffer_compat(dev,G31_IO_STREAM_IN_IF,arg);
            break;

        case G31_IOCTL_GET_ADC_LEVEL:
            status=g31_get_adc_level(dev,&word);

            if(likely(status==0))
            {
                status=put_user(word,(uint16_t __user*)compat_ptr(arg));
            }
            break;

        case G31_IOCTL_DDC1_0_SET_FREQUENCY:
            if(unlikely(get_user(dword,(uint32_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g31_ddc1_set_frequency(dev,0,dword);
            break;

        case G31_IOCTL_DDC1_0_START_COMPAT:
            if(unlikely(copy_from_user(&ddc1_start_params,(void __user*)compat_ptr(arg),sizeof(ddc1_start_params))!=0))
            {
                return -EFAULT;
            }

            status=g31_ddc1_start(dev,0,&ddc1_start_params.params,(void __user*)compat_ptr(ddc1_start_params.coeffs),ddc1_start_params.buffer_size,ddc1_start_params.bytes_per_sec);
            break;

        case G31_IOCTL_DDC1_0_STOP:
            g31_ddc1_stop(dev,0);
            status=0;
            break;

        case G31_IOCTL_IF_START:
            if(unlikely(get_user(word,(uint16_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g31_if_start(dev,word);
            break;

        case G31_IOCTL_IF_STOP:
            g31_if_stop(dev);

            status=0;
            break;

        case G31_IOCTL_GET_STATE:
            G31_DEV_LOCK;

            status=put_user(dev->state,(struct g31_state __user*)compat_ptr(arg));

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_SET_NOISE_BLANKER:
            if(unlikely(get_user(word,(uint16_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g31_set_noise_blanker(dev,word);
            break;

        case G31_IOCTL_DDC1_0_CANCEL_BUFFERS:
            g31_cancel_buffers(dev,G31_IO_STREAM_IN_DDC);
            status=0;
            break;

        case G31_IOCTL_USB_CONTROL1_READ_COMPAT:
            if(unlikely(copy_from_user(&usb_control1,(struct g31_usb_control1 __user*)compat_ptr(arg),sizeof(usb_control1))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control1_read(dev,usb_control1.params,(void __user*)compat_ptr(usb_control1.buffer),usb_control1.size,compat_ptr(usb_control1.transferred));

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL1_WRITE_COMPAT:
            if(unlikely(copy_from_user(&usb_control1,(struct g31_usb_control1 __user*)compat_ptr(arg),sizeof(usb_control1))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control1_write(dev,usb_control1.params,(void __user*)compat_ptr(usb_control1.buffer),usb_control1.size,compat_ptr(usb_control1.transferred));

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL2_READ_COMPAT:
            if(unlikely(copy_from_user(&usb_control2,(struct g31_usb_control2 __user*)compat_ptr(arg),sizeof(usb_control2))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control2_read(dev,(void __user*)compat_ptr(usb_control2.buffer),usb_control2.size,compat_ptr(usb_control2.transferred));

            G31_DEV_UNLOCK;
            break;

        case G31_IOCTL_USB_CONTROL2_WRITE_COMPAT:
            if(unlikely(copy_from_user(&usb_control2,(struct g31_usb_control2 __user*)compat_ptr(arg),sizeof(usb_control2))!=0))
            {
                return -EFAULT;
            }

            G31_DEV_LOCK;

            status=dev->func_usb_control2_write(dev,(void __user*)compat_ptr(usb_control2.buffer),usb_control2.size,compat_ptr(usb_control2.transferred));

            G31_DEV_UNLOCK;
            break; 
        case G31_IOCTL_SET_POWER:
            if(unlikely(get_user(byte,(uint8_t __user*)compat_ptr(arg))<0))
            {
                return -EFAULT;
            }

            status=g31_set_power(dev,byte,1);
            break;

        case G31_IOCTL_CONFIGURE_FPGA_COMPAT:
            if(unlikely(copy_from_user(&fpga,(struct g31_fpga_data __user*)compat_ptr(arg),sizeof(fpga))!=0))
            {
                return -EFAULT;
            }

            status=g31_configure_fpga(dev,(void __user*)compat_ptr(fpga.buffer),fpga.size);
            break;

        case G31_IOCTL_GET_CONFIG:
            status=copy_to_user((void __user*)compat_ptr(arg),dev->config,sizeof(dev->config))?-EFAULT:0;
            break;

        case G31_IOCTL_RF_CONTROL1:
            if(unlikely(copy_from_user(&rf,(struct g31_rf_control __user*)arg,sizeof(rf))!=0))
            {
                return -EFAULT;
            }

            status=g31_rf_control1(dev,&rf);
            break;
//*
        case G31_IOCTL_RF_CONTROL2:
            if(unlikely(copy_from_user(&rf2,(struct g31_rf_control2 __user*)arg,sizeof(rf2))!=0))
            {
                return -EFAULT;
            }

            status=g31_rf_control2(dev,&rf2);
            break;
//*/
        default:
            status=-EINVAL;
            break;
    }

    return status;
}
#endif

module_init(g31_init);
module_exit(g31_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("RADIXON");
MODULE_DESCRIPTION(G31_DESCRIPTION);
